!+
! Module: epatplot_mod
!
! Authors:  Michael Freyberg - MPE Garching
!           Frank Haberl     - MPE Garching
!           Konrad Dennerl   - MPE Garching
!           Hermann Brunner  - MPE Garching
!      
! Description: Plot EPIC pn and EPIC MOS event pattern statistics
!
! Comments:
!
!-

module epatplot_mod

  use eventAttributes
  use cal
  use caloaldefs
  use dal
  use param
  use errorHandling

  implicit none

  private

  public epatplot

contains

!-----------------------------------------------------------------------
subroutine epatplot
!-----------------------------------------------------------------------

  implicit none

  character(len=20)  ,parameter  :: CodeVersion = '1.20'
  integer(kind=Int32),parameter  :: maxbin = 4096, maxbin_times_2 = 2*maxbin
  integer(kind=Int32),parameter  :: maxpat = 4, maxpat_plus_1 = maxpat+1, & ! 5 = invalid
                                    maxpat_plus_2 = maxpat+2, &
                                    maxpat_plus_3 = maxpat+3, &
                                    maxpat_plus_4 = maxpat+4, &
                                    maxpat_plus_5 = maxpat+5
  integer(kind=Int32),parameter  :: maxmodel = maxpat
  integer                        :: statusPG, pgopen
!
  logical(kind=bool)  :: enable_qdp = .false., &
                         enable_pum = .false., &
                         enable_y   = .false.
!
  real(kind=single)   :: h, he, hs, hd, ht, hq, hv, hh, sigma, s,      &
                         xmin, xmax, ymin, ymax, xtxt, ytxt, yt0, yt1, &
                         xdat, ydat, xsiz, ysiz, srcposx, srcposy,     &
                         xamin, xamax, xasiz, backscal,                &
                         vxmin, vxmax, vymin, vymax, grey,             &
                         sprb(maxbin), rhis(maxpat,maxbin),            &
                         xhis(1:maxpat_plus_4,1:maxbin_times_2),       &
                         yhis(1:maxpat_plus_4,1:maxbin_times_2),       &
                         model(maxmodel,maxbin), livetime, y,          &
                         plotyrange(1:2),mod(maxpat),modnorm(maxpat)
  integer(kind=Int32) :: ntrow, ntrow_bkg, nsrow, i, j, k, npos, nxpos, mpos, fpos, &
                         nhas, mhas, om1, om2,                         &
                         ix0, ix, iy0, iy1, iy, iex, irow,             &
                         mbin, kbin, ibin, ks, kd, kt, kq, kv, kh, kpat, &
                         khis(1:maxpat_plus_5,1:maxbin), kkpat,        &
                         khis_bkg(1:maxpat_plus_5,1:maxbin),           &
                         jm(maxbin), jp(maxbin), knpat,                &
                         ipat, jpat, jcts(maxbin),                     &
                         icol(maxpat_plus_4), k1, k2, nbin(maxpat_plus_4)
  integer(kind=Int32) :: CALstat_mode, CALstat_node, CALstat_ccd, &
                         CALstat_inst
  integer(kind=Int16) :: ix16, iy16
  integer(kind=Int32), dimension(2)  :: imageDim
  integer, parameter  :: nccdlimits = 6
  integer             :: ccdlimits(1:nccdlimits),ccdmin,ccdmax,   &
                         rawxmin,rawxmax,rawymin,rawymax
! integer             :: dthresh, mthresh(0:5)
  integer             :: iw, ierr, mmp, istart, iend, nsel,       &
                         calMode, userMode, node = 0, revolut,    &
                         nccd = 12, iccd, quadrant, ccdid,        &
                         AccessType = READ, EventsFileType,       &
                         xw1 = 1, xw2 = 1, nx, ndx,               &
                         yw1 = 1, yw2 = 1, ny, ndy
  logical(kind=bool)  :: is_ps = .false., is_gif = .false.,       &
                         flag0 = .true., hasCCD(1:12,0:1),        &
                         is_epn = .false., is_mos = .false.,      &
                         is_rgs = .false., ccdlimsel = .true.
  logical(kind=bool)  :: xaxisadu = .false., useplotfile = .false.,    &
                         withflag = .true., withoutputmask = .false.,  &
                         withdetxy = .false., withsrcxy = .true. ,     &
                         badpix_from_events = .true.,                  &
                         offsets_from_events = .true.,                 &
                         hasBADPIX(0:12,0:1) = .false.,                &
                         hasOFFSET(0:12,0:1) = .false.,                &
                         withccdnr = .true., singleChip = .false. ,    &
                         lunrecogn = .false., hasallcolumns = .false., &
                         hasallcolumns_bkg = .false.,                  &
                         hasdetx = .true., hasdety = .true.,           &
                         hasrawx = .true., hasrawy = .true.,           &
                         hasrawx_bkg = .true., hasrawy_bkg = .true.,   &
                         compressed, withbackgroundset = .false.,      &
                         nosignal=.false., modifyInSet = .true.,       &
                         linvalid3 = .false., linvalid4 = .false.,     &
                         levattgood = .true.

  logical(kind=bool)  :: emptyEvents = .false.

  character(len=1)    :: chpt(maxpat_plus_4)
  character(len=4)    :: ccdmode, mmode(0:7),mosMode(0:7)
  character(len=8)    :: blockname(0:12),blockname2(0:12), revlab
  character(len=40)   :: string, instmode, device, ext, xlabel, &
                         instrument, insttxt, rawytxt, sdfractxt, filter, &
                         rawxtxt, detxtxt, detytxt, srcxtxt, srcytxt
  character(len=1024) :: infilename, bkgfilename, outfilename,        &
                         canonicalname,                               &
                         outdir, plotfile, outqdpname, outmaskname,   &
                         omsk1, omsk2
  logical             :: special, special_1, special_2, special_3
  character(len=2)    :: inst2
  character(len=22)   :: odf_icd
  character(len=40)   :: expidstr, obs_id, object
!
  integer(kind=int16),dimension(:,:), pointer     :: FOVmap
  integer(kind=Int8), dimension(:,:), pointer     :: map_arrayP
  integer(kind=Int8), dimension(:,:), allocatable :: map_array

  type(DataSetT)      :: infile, bkgfile, outmask
  type(ArrayT)        :: map
  type(TableT)        :: events, bkgevents, badpix, offset
  type(ColumnT)       :: pattern,pi,flag,rawx,rawy,ccdnr,yextent,detx,dety
  type(ColumnT)       :: pattern_bkg,pi_bkg,flag_bkg,rawx_bkg,rawy_bkg, &
                         ccdnr_bkg
  type(PatternFractionServerT)    :: pfds
! type(CurveT)                    :: curve
  integer(kind=Int8),  dimension(:), pointer :: patternCol,ccdnrCol
  integer(kind=Int8),  dimension(:), pointer :: patternCol_bkg,ccdnrCol_bkg
  integer(kind=Int16), dimension(:), pointer :: piCol,rawxCol,rawyCol, &
                                                yextentCol, detxCol, detyCol
  integer(kind=Int16), dimension(:), pointer :: piCol_bkg,rawxCol_bkg, &
                                                rawyCol_bkg
  integer(kind=Int32), dimension(:), pointer :: flagCol, flagCol_bkg
  real(kind=double), dimension(:), pointer   :: pattFracChan
!
  real(kind=single)           ::  conv, rconv, userY = 190.0
  real(kind=single),parameter ::  xpos_epn =  32., xpos_mos=300.
  real(kind=single),parameter ::  ypos_epn = 190., ypos_mos=300.
  real(kind=single)           ::  xpos, ypos, srcpos, xavpos, xdavpos, ydavpos
  real(kind=single),parameter ::  chan = 15.0 ! eV per MOS bin
  real(kind=single),parameter ::  econv_epn = 5.0, econv_mos = 3.4
  real(kind=single)           ::  econv       ! eV per adu channel
!
  integer(kind=int8),parameter :: nband = 6
  real(kind=single)           ::  kband(0:nband,1:maxpat_plus_5), &
                                  kband_bkg(0:nband,1:maxpat_plus_5)
  real(kind=single)           ::  kfrac(maxpat),kerror(maxpat)
  real(kind=single)           ::  e_low(0:nband),e_hig(0:nband), &
                                  e_sum(0:nband,1:maxpat_plus_5),&
                                  e_sum_bkg(0:nband,1:maxpat_plus_5), &
                                  pileuprange(2)

  data e_low    / 0500,200,0500,2000,4500,07500,01000 /
  data e_hig    / 2000,500,2000,4500,7500,12000,32800 /
! data icol     / 2, 4, 3, 5, 15 /   ! 18 -> 15
  data icol     / 2, 4, 3, 5, 15, 7, 12, 8 /
!                 2 = red
!                 3 = light green
!                 4 = dark blue
!                 5 = light blue (turquoise)
!                 6 = really pink
!                 7 = yellow
!                 8 = orange
!                12 = purple
!                13 = dark red
!                14 = dark grey
!                15 = light grey
!                 1 = black
!                 0 = white
! data mthresh  / 40,40,40,40,80,46 /
  data mMode    / 'FF','eFF','LW','SW','TI','BU','  ','  ' /
  data mosMode  / 'FF','RFS','W2','W3','W4','W5','W6','FU' /
! data chpt     / 's', 'd', 't', 'q', 'u' /
! data chpt     / 's', 'd', 't', 'q', '3' , '4'/
  data chpt     / 's', 'd', 't', 'q', 'u' , 'x', 'v', 'h'/
!
!-----------------------------------------------------------------------
! start the program:
!-----------------------------------------------------------------------
!
  write(errstr,*) 'epatplot ',trim(CodeVersion),' is running...'
  call message(AppMsg,Sparse)

  linvalid3 = .false.
  linvalid4 = .false.
  levattgood = .true.

! -------------------------------------------
! initialize the histograms:
! -------------------------------------------

  khis(1:maxpat_plus_5,1:maxbin) = 0
  xhis(1:maxpat_plus_4,1:maxbin_times_2) = 0
  yhis(1:maxpat_plus_4,1:maxbin_times_2) = 0
  rhis(1:maxpat,1:maxbin) = 0.0
  model(1:maxmodel,1:maxbin) = 0.0
  jm(1:maxbin) = 0
  jp(1:maxbin) = 0
  jcts(1:maxbin) = 0
  e_sum(0:nband,1:maxpat_plus_5) = 0

! -------------------------------------------
! get input parameters:
! -------------------------------------------
  call getInParamValues(infilename, sigma, device, outdir, withflag, &
                        useplotfile, plotfile, xaxisadu, enable_qdp, &
                        enable_pum, userMode, enable_y, userY,       &
                        withoutputmask, outmaskname, ccdlimits,      &
                        nccdlimits, plotyrange, withbackgroundset,   &
                        bkgfilename, backscal, modifyInSet, pileuprange, &
                        withdetxy,withsrcxy)

  rawxmin = ccdlimits(1)
  rawxmax = ccdlimits(2)
  rawymin = ccdlimits(3)
  rawymax = ccdlimits(4)
  ccdmin  = ccdlimits(5)
  ccdmax  = ccdlimits(6)

  e_low(0) = pileuprange(1)
  e_hig(0) = pileuprange(2)

  call stripSlash(infilename)
  call getCanonicalName(infilename, canonicalname)
  call stripSlash(canonicalname)
  call stripSlash(plotfile)

  if (withoutputmask) then
    om1 = index(outmaskname,'##')
    om2 = om1+2
    write(errstr,'(1x,a,t35,i9)') 'CCD placeholder ## at position: ',om1
    call message(AppMsg,Noisy)
    if (om1>1) then
      omsk1 = outmaskname(1:om1-1)
      omsk2 = outmaskname(om2:)
    elseif (om1 == 1) then
      omsk1 = ""
      omsk2 = outmaskname(om2:)
    else
      omsk1 = outmaskname
      omsk2 = ""
    endif
  endif

  select case (device)
    case ('/VCPS')
      ext = 'ps'
    case ('/GIF')
      ext = 'gif'
    case default
      ext = 'ps'
  end select

  if (useplotfile) then
    outfilename = trim(outdir) // '/' // trim(plotfile)
    outqdpname  = trim(outdir) // '/' // trim(canonicalname) // '.qdp'
  else
    outfilename = trim(outdir) // '/' // trim(canonicalname) // '.' // trim(ext)
    outqdpname  = trim(outdir) // '/' // trim(canonicalname) // '.qdp'
  endif

  call stripSlash(outfilename)
  call stripSlash(outqdpname)

! print*,'|',trim(outfilename),'|'

  write(errstr,'(1x,a,a)') 'Accessing input data set: ',trim(infilename)
  call message(AppMsg,Verbose)

  if (.not.setexists(infilename)) then
    call error("noSet","Input data set not found")
  endif

  compressed = (index(infilename,".gz" )>0 .or. index(infilename,".GZ" )>0 .or. &
                index(infilename,".ftz")>0 .or. index(infilename,".FTZ")>0)

  if ( (.not.compressed) .and. (.not. modifyInSet)) compressed = .true.

  if (.not.compressed) then
    infile = dataSet(infilename,MODIFY,LOW_MEMORY)
  else
    infile = dataSet(infilename,READ,LOW_MEMORY)
  end if

  if (.not. hasBlock(infile,'EVENTS')) then    ! no EVENTS extension
    call error("noEvents","No EVENTS extension in input data set")
  endif
  if (.not.compressed) then
    events = table(infile,'EVENTS',MODIFY)
  else
    events = table(infile,'EVENTS',READ)
  endif


  if (withBackgroundSet) then
    write(errstr,'(1x,a,a)') 'Accessing input data set: ',trim(bkgfilename)
    call message(AppMsg,Verbose)

    if (.not.setexists(bkgfilename)) then
      call error("noSet","Background data set not found")
    endif
    bkgfile = dataSet(bkgfilename,READ,LOW_MEMORY)

    if (.not. hasBlock(bkgfile,'EVENTS')) then    ! no EVENTS extension
      call error("noEvents","No EVENTS extension in background data set")
    endif
    bkgEvents = table(bkgfile,'EVENTS',READ)
  endif

!
! read FILTER attribute and fill relevant text string
!
  if (hasAttribute(infile,"FILTER")) then
    filter = trim(stringAttribute(infile,"FILTER"))
    write(errstr,'(1x,a,t39,a)') 'Filter:     ',trim(filter)
    call message(AppMsg,Verbose)
  else
    filter = ' '
    call message(AppMsg,Verbose,'FILTER attribute not found')
  endif

!
! read OBJECT attribute and fill relevant text string,
! the standard PPS OBJECT length is only 18 characters
!
  if (hasAttribute(infile,"OBJECT")) then
    object = trim(stringAttribute(infile,"OBJECT"))
    write(errstr,'(1x,a,t39,a)') 'Object:     ',trim(object)
    call message(AppMsg,Verbose)
  else
    object = ' '
    call message(AppMsg,Verbose,'OBJECT attribute not found')
  endif

!
! read INSTRUME attribute and fill relevant text string
!
  if (hasAttribute(events,"INSTRUME")) then
    instrument = trim(stringAttribute(events,"INSTRUME"))
    write(errstr,'(1x,a,t39,a)') 'Instrument: ',trim(instrument)
    call message(AppMsg,Verbose)
  else
    instrument = ' '
    call message(AppMsg,Verbose,'INSTRUME attribute not found')
  endif

  select case (instrument)
    case ('EPN')
      insttxt = 'EPIC pn'
      is_epn = .true.
      CALstat_inst = EPN
      inst2 = 'PN'
    case ('EMOS1')
      insttxt = 'EPIC MOS1'
      is_mos = .true.
      CALstat_inst = EMOS1
      inst2 = 'M1'
    case ('EMOS2')
      insttxt = 'EPIC MOS2'
      is_mos = .true.
      CALstat_inst = EMOS2
      inst2 = 'M2'
    case ('RGS1')
      insttxt = 'RGS1'
      nccd = 9
      is_rgs = .true.
      CALstat_inst = RGS1
      inst2 = 'R1'
    case ('RGS2')
      insttxt = 'RGS2'
      nccd = 9
      is_rgs = .true.
      CALstat_inst = RGS2
      inst2 = 'R2'
    case default
      insttxt = 'EPIC pn'
      is_epn = .true.
      CALstat_inst = EPN
      inst2 = 'PN'
  end select

  if (.not.is_epn.and..not.is_mos) then
    if (xaxisadu) then
      call warning("isnotEPN","xaxisadu=Y only implemented for EPIC.")
      xaxisadu = .false.
    endif
  endif

  if (is_mos) then
    econv=econv_mos ! channel -> eV
  else
    econv=econv_epn ! channel -> eV
  end if

  if (xaxisadu) then
    conv = 1.0
    rconv = econv
    if (is_mos) then
      xlabel = 'ADU Channel'
    else
      xlabel = 'ADU Channel [5 eV]'
    endif
  else
    conv = econv
    rconv = 1.0
    xlabel = 'PI Channel [eV]'
  endif

!
! get PI energy of ADU channel
!
  if (is_epn) then
    nccd = 12
    ndx = EPN_MAX_X
    ndy = EPN_MAX_Y
    yt0 = 0.1
    yt1 = 0.2
  elseif (is_mos) then
    nccd = 7
    ndx = EMOS_MAX_X
    ndy = EMOS_MAX_Y
    mMode=mosMode
    yt0 = 0.125
    yt1 = 0.25
  else
    call error("isnotEPIC","only EPIC model curves are implemented yet")
    withoutputmask = .false.   ! only implemented for EPIC at the moment
    ndx = 1
    ndy = 1
    yt0 = 0.125
    yt1 = 0.25
  endif
  imageDim(1) = ndx
  imageDim(2) = ndy
  nx =  ndx
  ny =  ndy

!
! access events file columns
!
  ntrow = numberOfRows(events)
  if (ntrow .le. 0) then
    emptyEvents = .true.
!   call error('emptyInput','No events in input file.') ! version 1.19 and earlier
    call warning("emptyEvents","empty EVENTS extension, no data for plotting")
    goto 990
  else
    emptyEvents = .false.
    write(errstr,'(1x,a,t35,i9)') 'Number of events in input file: ',ntrow
    call message(AppMsg,Verbose)
  endif

  if (.not.hasColumn(events,'PATTERN')) then
    call error("noPattern","No PATTERN column in input data set")
  endif
  pattern = column(events,'PATTERN',READ)
  patternCol => int8data(pattern)

  if (.not.hasColumn(events,'PI')) then
    call error("noPI","No PI column in input data set")
  endif
  pi = column(events,'PI',READ)
  piCol => int16data(pi)

  if (.not.hasColumn(events,'RAWX')) then
    call warning("noRAWX","No RAWX column in input data set")
    hasrawx=.false.
  else
    rawx = column(events,'RAWX',READ)
    rawxCol => int16data(rawx)
    hasrawx=.true.
  endif

  if (.not.hasColumn(events,'RAWY')) then
    call warning("noRAWY","No RAWY column in input data set")
    hasrawy=.false.
  else
    rawy = column(events,'RAWY',READ)
    rawyCol => int16data(rawy)
    hasrawy=.true.
  endif

  if (.not.hasColumn(events,'DETX')) then
    call warning("noDETX","No DETX column in input data set")
    hasdetx=.false.
  else
    detx = column(events,'DETX',READ)
    detxCol => int16data(detx)
    hasdetx=.true.
  endif

  if (.not.hasColumn(events,'DETY')) then
    call warning("noDETY","No DETY column in input data set")
    hasdety=.false.
  else
    dety = column(events,'DETY',READ)
    detyCol => int16data(dety)
    hasdety=.true.
  endif

  if (.not.hasColumn(events,'FLAG')) then
    if (withflag) then
      call warning("noFlag", &
          "No FLAG column in input data set, using all events for plots.")
      flag0 = .true.
      withflag = .false.
    endif
  else
    flag = column(events,'FLAG',READ)
    flagCol => int32data(flag)
  endif

  if (.not.hasColumn(events,'CCDNR')) then
    if (withccdnr) then
      call warning("noCCDnr", &
          "No CCDNR column in input data set, will check all CCDs.")
      withccdnr = .false.
    endif
  else
    ccdnr = column(events,'CCDNR',READ)
    ccdnrCol => int8data(ccdnr)
  endif

  if (hasAttribute(events,"SUBMODE")) then
    instmode = trim(stringAttribute(events,"SUBMODE"))
  else
    instmode = ' '
    call message(AppMsg,Verbose,'SUBMODE attribute not found')
  endif

  if (hasAttribute(events,"LIVETIME")) then
    livetime = real64Attribute(events,"LIVETIME")
    write(errstr,'(1x,a,t35,f9.1,a)') 'Livetime by evselect: ',livetime,' [s]'
    call message(AppMsg,Verbose)
  else
    livetime = 0.0
  endif

  if (hasAttribute(events,"REVOLUT")) then
    revolut = int16Attribute(events,"REVOLUT")
    write(revlab,'(a,i4.4)') 'Rev.',revolut
  else
    revolut = 0
    revlab = ' '
  endif

  if (hasAttribute(events,"OBS_ID")) then
    obs_id = trim(stringAttribute(events,"OBS_ID"))
  else
    obs_id = '0000000000'
    call message(AppMsg,Noisy,'OBS_ID attribute not found')
  endif

  if (hasAttribute(events,"EXPIDSTR")) then
    expidstr = trim(stringAttribute(events,"EXPIDSTR"))
  else
    expidstr = 'X000'
    call message(AppMsg,Noisy,'EXPIDSTR attribute not found')
  endif

  write(ODF_icd,'(i4.4,a1,a10,a1,a2,a4)') revolut,'_',trim(obs_id), &
       '_',trim(inst2),trim(expidstr)

  special_1 = .false.
  special_2 = .false.
  special_3 = .false.
  if (trim(ODF_icd) == '0056_0122330701_PNU002') then
    special_1 = .true.
  elseif (trim(ODF_icd) == '0301_0029740101_PNS303') then
    special_1 = .true.
  endif
  special = (special_1 .or. special_2 .or. special_3)

  write(errstr,*) 'ODF ',trim(ODF_icd),'   ',special_1,special_2,special_3
  call message(AppMsg,Verbose)
  node = 0

  if (withoutputmask) allocate(map_array(1:ndx,1:ndy))
!
! access bkg events file columns
!
  if (withbackgroundset) then
    ntrow_bkg = numberOfRows(bkgevents)
    if (ntrow_bkg .le. 0) then
      call error('emptyInput','No events in background file.')
    else
      write(errstr,'(1x,a,t35,i9)') &
        'Number of events in background file: ',ntrow_bkg
      call message(AppMsg,Verbose)
    endif

    if (.not.hasColumn(bkgevents,'PATTERN')) then
      call error("noPattern","No PATTERN column in background data set")
    endif
    pattern_bkg = column(bkgevents,'PATTERN',READ)
    patternCol_bkg => int8data(pattern_bkg)

    if (.not.hasColumn(bkgevents,'PI')) then
      call error("noPI","No PI column in background data set")
    endif
    pi_bkg = column(bkgevents,'PI',READ)
    piCol_bkg => int16data(pi_bkg)

    if (.not.hasColumn(bkgevents,'RAWY')) then
      call warning("noRAWY","No RAWY column in background data set")
      hasrawy_bkg=.false.
    else
      rawy_bkg = column(bkgevents,'RAWY',READ)
      rawyCol_bkg => int16data(rawy_bkg)
      hasrawy_bkg=.true.
    endif

    if (.not.hasColumn(bkgevents,'RAWX')) then
      call warning("noRAWX","No RAWX column in background data set")
      hasrawx_bkg=.false.
    else
      rawx_bkg = column(bkgevents,'RAWX',READ)
      rawxCol_bkg => int16data(rawx_bkg)
      hasrawx=.true.
    endif

    if (.not.hasColumn(bkgevents,'FLAG')) then
      if (withflag) then
        call warning("noFlag", &
        "No FLAG column in background data set, using all events for plots.")
        flag0 = .true.
        withflag = .false.
      endif
    else
      flag_bkg = column(bkgevents,'FLAG',READ)
      flagCol_bkg => int32data(flag_bkg)
    endif

    if (.not.hasColumn(bkgevents,'CCDNR')) then
      if (withccdnr) then
        call warning("noCCDnr", &
          "No CCDNR column in background data set, will check all CCDs.")
        withccdnr = .false.
      endif
    else
      ccdnr_bkg = column(bkgevents,'CCDNR',READ)
      ccdnrCol_bkg => int8data(ccdnr_bkg)
    endif
  end if

  select case (instmode)
! PN modes
    case ('PrimeFullWindow')
      calMode = 0
      CALstat_Mode = PRIME_FULL_WINDOW
    case ('PrimeFullWindowExtended')
      calMode = 1
      CALstat_Mode = PRIME_FULL_WINDOW_EXTENDED
    case ('PrimeLargeWindow')
      calMode = 2
      CALstat_Mode = PRIME_LARGE_WINDOW
    case ('PrimeSmallWindow')
      calMode = 3
      CALstat_Mode = PRIME_SMALL_WINDOW
    case ('FastTiming')
      calMode = 4
      CALstat_Mode = FAST_TIMING
    case ('FastBurst')
      calMode = 5
      CALstat_Mode = FAST_BURST
! additional MOS modes
    case ('PrimePartialRFS')
      CALstat_Mode = PRIME_PARTIAL_RFS
      calMode = CALstat_Mode
    case ('PrimePartialW2')
      CALstat_Mode = PRIME_PARTIAL_W2
      calMode = CALstat_Mode
    case ('PrimePartialW3')
      CALstat_Mode = PRIME_PARTIAL_W3
      calMode = CALstat_Mode
    case ('PrimePartialW4')
      CALstat_Mode = PRIME_PARTIAL_W4
      calMode = CALstat_Mode
    case ('PrimePartialW5')
      CALstat_Mode = PRIME_PARTIAL_W5
      calMode = CALstat_Mode
    case ('PrimePartialW6')
      CALstat_Mode = PRIME_PARTIAL_W6
      calMode = CALstat_Mode
    case ('FastUncompressed')
      CALstat_Mode = FAST_UNCOMPRESSED
      calMode = CALstat_Mode
    case ('FastCompressed')
      CALstat_Mode = FAST_COMPRESSED
      calMode = CALstat_Mode
! RGS modes
    case ('Spectroscopy')
      CALstat_Mode = SPECTROSCOPY
      calMode = CALstat_Mode
    case ('HighEventRate')
      CALstat_Mode = HIGH_EVENT_RATE
      calMode = CALstat_Mode
    case ('HighEventRateWithSES')
      CALstat_Mode = HIGH_EVENT_RATE_WITH_SES
      calMode = CALstat_Mode
    case ('HighEventRateWithSER')
      CALstat_Mode = HIGH_EVENT_RATE_WITH_SER
      calMode = CALstat_Mode
    case ('HighTimeResolutionSingle')
      CALstat_Mode = HIGH_TIME_RESOLUTION_SINGLE
      calMode = CALstat_Mode
    case ('HighTimeResolutionMultiple')
      CALstat_Mode = HIGH_TIME_RESOLUTION_MULTIPLE
      calMode = CALstat_Mode
    case default
      ccdmode = '  '
      calMode = 0
      CALstat_Mode = PRIME_FULL_WINDOW
!     dthresh = 40
  end select

  if (enable_pum) then
    if (is_epn) then
      if (userMode.le.5) calMode = userMode
      select case (calMode)
        case (0)
          CALstat_Mode = PRIME_FULL_WINDOW
        case (1)
          CALstat_Mode = PRIME_FULL_WINDOW_EXTENDED
        case (2)
          CALstat_Mode = PRIME_LARGE_WINDOW
        case (3)
          CALstat_Mode = PRIME_SMALL_WINDOW
        case (4)
          CALstat_Mode = FAST_TIMING
        case (5)
          CALstat_Mode = FAST_BURST
        case default
          CALstat_Mode = PRIME_FULL_WINDOW
          calMode=0
          call warning("InvalidUserMode","Usermode>5 not supported for PN")
      end select
    else if (is_mos) then
      calMode = userMode
      Calstat_Mode = calMode
    else
      call error("isnotEPIC","Model curves are available only for EPIC")
    end if
  end if   
!  dthresh = mthresh(calMode)
  ccdmode = mmode(calMode)
  write(errstr,'(1x,a,t35,a)') 'Instrument mode of observation:',trim(instmode)
  call message(AppMsg,Verbose)
  write(errstr,'(1x,a,t35,f9.1)') 'Sigma for rebinning: ',sigma
  call message(AppMsg,Verbose)

! +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

! first the background

  if (withbackgroundset) then
    call message(AppMsg,Noisy,'Reading background data set ...')
    hasallcolumns_bkg = (hasrawx_bkg .and. hasrawy_bkg .and. withccdnr)

    ypos = 0.
    xavpos = 0.
    npos = 0
    nxpos = 0
    mpos = 0
    fpos = 0
    knpat = 0 ! doubles
    ccdlimsel = .true.
    hasCCD(1:12,0:1) = .false.
    khis_bkg(1:,1:maxbin) = 0
    do i = 1, ntrow_bkg
      if (withccdnr) then
        iccd = ccdnrCol_bkg(i-1)
        if ((iccd.ge.1).and.(iccd.le.nccd)) then
          hasCCD(iccd,node) = .true.
        endif
      endif
      jpat = patternCol_bkg(i-1)
      if (withflag) flag0 = (flagCol_bkg(i-1) == 0)
      if (is_EPN) then
        lunrecogn = &
        ((((iand(flagCol_bkg(i-1),EVATT_INVALID_PATTERN)>0).and.(jpat>4)) .or. &
                (jpat>12)) .and. piCol_bkg(i-1)>0 .and. &
                (iand(flagCol_bkg(i-1),EVATT_CLOSE_TO_CCD_WINDOW)==0) .and. &
                (iand(flagCol_bkg(i-1),EVATT_CLOSE_TO_ONBOARD_BADPIX)==0) .and. &
                (iand(flagCol_bkg(i-1),EVATT_CLOSE_TO_BRIGHTPIX)==0) .and. &
                (iand(flagCol_bkg(i-1),EVATT_CLOSE_TO_DEADPIX)==0) .and. &
                (iand(flagCol_bkg(i-1),EVATT_COSMIC_RAY)==0) .and. &
                (iand(flagCol_bkg(i-1),EVATT_TRAILING)==0) .and. &
                (iand(flagCol_bkg(i-1),EVATT_ON_BADPIX)==0) .and. &
                (iand(flagCol_bkg(i-1),EVATT_SECONDARY)==0) .and. &
                (iand(flagCol_bkg(i-1),EVATT_MIP_ASSOCIATED)==0) .and. &
                (iand(flagCol_bkg(i-1),EVATT_OUT_OF_CCD_WINDOW)==0) .and. &
                (iand(flagCol_bkg(i-1),EVATT_OUT_OF_FOV)==0) .and. &
                (iand(flagCol_bkg(i-1),EVATT_IN_BAD_FRAME)==0) .and. &
                (iand(flagCol_bkg(i-1),EVATT_IN_SPOILED_FRAME)==0))
        if (hasallcolumns_bkg) ccdlimsel = (rawxCol_bkg(i-1).ge.rawxmin .AND. &
                                        rawxCol_bkg(i-1).le.rawxmax .AND. &
                                        rawyCol_bkg(i-1).ge.rawymin .AND. &
                                        rawyCol_bkg(i-1).le.rawymax .AND. &
                                        ccdnrCol_bkg(i-1).ge.ccdmin .AND. &
                                        ccdnrCol_bkg(i-1).le.ccdmax)
      elseif (is_MOS) then
        lunrecogn = (jpat.eq.31 .and. flag0)
      else
        lunrecogn = .FALSE.
      endif

      kkpat = 0
      if (jpat.eq.0 .and. flag0 .and. ccdlimsel) then                   ! singles
        kpat = 1
      elseif (jpat.ge.1.and.jpat.le.4 .and. flag0 .and. ccdlimsel) then ! doubles
        kpat = 2
        if (jpat.eq.1.or.jpat.eq.3) then ! doubles vertical U+D
          kkpat = 7
          knpat = knpat+1
        elseif (jpat.eq.2.or.jpat.eq.4) then ! doubles horizontal R+L
          kkpat = 8
          knpat = knpat+1
        else
          print*,'ERROR'
        endif
      elseif (jpat.ge.5.and.jpat.le.8 .and. flag0 .and. ccdlimsel) then ! triples
        kpat = 3
      elseif (jpat.ge.9.and.jpat.le.12 .and. flag0 .and. ccdlimsel) then ! quadruples
        kpat = 4
      elseif (lunrecogn .and. ccdlimsel) then
        kpat = 5
        mpos = mpos+1
        if (verbosity(AppMsg) >= Noisy) then
          if (piCol_bkg(i-1)<220) then
            if (hasrawy_bkg.and.hasrawx_bkg) then
              write(errstr,'(i9,i5,i12,i6,i3,i4)') &
                  i,jpat,flagCol_bkg(i-1),piCol_bkg(i-1), &
                  rawxcol_bkg(i-1),rawycol_bkg(i-1)
              call message(AppMsg,Sparse)
            elseif (hasrawx_bkg) then
              write(errstr,'(i9,i5,i12,i6,i3)') &
                  i,jpat,flagCol_bkg(i-1),piCol_bkg(i-1), &
                  rawxcol_bkg(i-1)
              call message(AppMsg,Sparse)
            endif
          endif
        endif
      else
        kpat=6
      endif

      if (hasrawy_bkg .and. jpat.ge.0 .and. jpat.le.12.and. flag0 .and. ccdlimsel) then
        ypos = ypos + rawyCol_bkg(i-1)
        npos = npos+1
      end if
      if (hasrawx_bkg .and. jpat.ge.0 .and. jpat.le.12.and. flag0 .and. ccdlimsel) then
        xavpos = xavpos + rawxCol_bkg(i-1)
        nxpos = nxpos+1
      end if
!
! accumulate pattern types and energies of pattern types in bands
!
      if (kpat .le. maxpat_plus_1) then
        do j = 0,nband
          if (piCol_bkg(i-1).ge.e_low(j).and.piCol_bkg(i-1).lt.e_hig(j)) then
            e_sum_bkg(j,kpat) = e_sum_bkg(j,kpat) + piCol_bkg(i-1)
            kband_bkg(j,kpat) = kband_bkg(j,kpat) + 1.0
          endif
        enddo

      endif
!
! find the energy bin corresponding to event PI channel
!
      kbin = piCol_bkg(i-1)/econv

      if (kpat.ge.1 .and. kpat.le.maxpat_plus_1 .and.   &
          kbin.ge.1 .and. kbin.le.maxbin) then
        khis_bkg(kpat,kbin) = khis_bkg(kpat,kbin) + 1
      endif
!
! doubles HORI + VERT in addition separately
!
      if (kkpat .gt. 0 .and.   &
          kbin.ge.1 .and. kbin.le.maxbin) then
        khis_bkg(kkpat,kbin) = khis_bkg(kkpat,kbin) + 1
      endif
    enddo

    if (withflag) call release(flag_bkg)
    if (withccdnr) call release(ccdnr_bkg)
    call release(pattern_bkg)
    call release(pi_bkg)
    if (hasrawx) call release(rawx_bkg)
    if (hasrawy) call release(rawy_bkg)
    call release(bkgevents)
    call release(bkgfile)

    write(errstr,'(1x,a,t35,12l1)') 'CCDs in background file: ', &
                  (hasCCD(iccd,node),iccd=1,nccd)
    call message(AppMsg,Verbose)

    write(errstr,'(1x,a,t35,i9  )') &
      'Number of recognized background photons: ',npos
    call message(AppMsg,Verbose)
    write(errstr,'(1x,a,t35,i9  )') &
      'Number of unrecognized background cases: ',mpos
    call message(AppMsg,Verbose)
  end if

! +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

  call message(AppMsg,Noisy,'Reading input data set ...')

  hasallcolumns = (hasrawx .and. hasrawy .and. withccdnr)

  ypos = 0.
  xavpos = 0.
  npos = 0
  nxpos = 0
  mpos = 0
  fpos = 0
  knpat = 0 ! doubles
  ccdlimsel = .true.
  hasCCD(1:12,0:1) = .false.
  do i = 1, ntrow
    if (withccdnr) then
      iccd = ccdnrCol(i-1)
      if ((iccd.ge.1).and.(iccd.le.nccd)) then
        hasCCD(iccd,node) = .true.
      endif
    endif
    jpat = patternCol(i-1)
    if (withflag) flag0 = (flagCol(i-1) == 0)
    if (is_EPN) then
      lunrecogn = ((((iand(flagCol(i-1),EVATT_INVALID_PATTERN)>0).and.(jpat>4)) .or. &
                  (jpat>12)) .and. piCol(i-1)>0 .and. &
                  (iand(flagCol(i-1),EVATT_CLOSE_TO_CCD_WINDOW)==0) .and. &
                  (iand(flagCol(i-1),EVATT_CLOSE_TO_ONBOARD_BADPIX)==0) .and. &
                  (iand(flagCol(i-1),EVATT_CLOSE_TO_BRIGHTPIX)==0) .and. &
                  (iand(flagCol(i-1),EVATT_CLOSE_TO_DEADPIX)==0) .and. &
                  (iand(flagCol(i-1),EVATT_COSMIC_RAY)==0) .and. &
                  (iand(flagCol(i-1),EVATT_TRAILING)==0) .and. &
                  (iand(flagCol(i-1),EVATT_ON_BADPIX)==0) .and. &
                  (iand(flagCol(i-1),EVATT_SECONDARY)==0) .and. &
                  (iand(flagCol(i-1),EVATT_MIP_ASSOCIATED)==0) .and. &
                  (iand(flagCol(i-1),EVATT_OUT_OF_CCD_WINDOW)==0) .and. &
                  (iand(flagCol(i-1),EVATT_OUT_OF_FOV)==0) .and. &
                  (iand(flagCol(i-1),EVATT_IN_BAD_FRAME)==0) .and. &
                  (iand(flagCol(i-1),EVATT_IN_SPOILED_FRAME)==0))
      if (hasallcolumns) ccdlimsel = (rawxCol(i-1).ge.rawxmin .AND. &
                                      rawxCol(i-1).le.rawxmax .AND. &
                                      rawyCol(i-1).ge.rawymin .AND. &
                                      rawyCol(i-1).le.rawymax .AND. &
                                      ccdnrCol(i-1).ge.ccdmin .AND. &
                                      ccdnrCol(i-1).le.ccdmax)
    elseif (is_MOS) then
      lunrecogn = (jpat.eq.31 .and. flag0)
    else
      lunrecogn = .FALSE.
    endif

    kkpat = 0
    if (jpat.eq.0 .and. flag0 .and. ccdlimsel) then                    ! singles
      kpat = 1
    elseif (jpat.ge.1.and.jpat.le.4 .and. flag0 .and. ccdlimsel) then  ! doubles
      kpat = 2
      if (jpat.eq.1.or.jpat.eq.3) then ! doubles
        kkpat = 7
        knpat = knpat+1
      elseif (jpat.eq.2.or.jpat.eq.4) then ! doubles
        kkpat = 8
        knpat = knpat+1
      else
        print*,'ERROR'
      endif
    elseif (jpat.ge.5.and.jpat.le.8 .and. flag0 .and. ccdlimsel) then  ! triples
      kpat = 3
    elseif (jpat.ge.9.and.jpat.le.12 .and. flag0 .and. ccdlimsel) then ! quadruples
      kpat = 4
!   elseif (jpat.le.12 .and. .not.flag0) then    ! not FLAG=0
!   elseif (rawyCol(i-1).eq.200) then    !
!     kpat = 6
!     fpos = fpos+1  
!   else
!   elseif ((iand(flagCol(i-1),EVATT_INVALID_PATTERN)>0).or.(jpat>12)) then
    elseif (lunrecogn .and. ccdlimsel) then
      kpat = 5
      mpos = mpos+1
      if (verbosity(AppMsg) >= Noisy) then
        if (piCol(i-1)<220) then
          if (hasrawy.and.hasrawx) then
            write(errstr,'(i9,i5,i12,i6,i3,i4)') &
                  i,jpat,flagCol(i-1),piCol(i-1), &
                  rawxcol(i-1),rawycol(i-1)
            call message(AppMsg,Sparse)
          elseif (hasrawx) then
            write(errstr,'(i9,i5,i12,i6,i3)') &
                  i,jpat,flagCol(i-1),piCol(i-1), &
                  rawxcol(i-1)
            call message(AppMsg,Sparse)
          endif
        endif
      endif
    else
      kpat=6
    endif

    if (hasrawy .and. jpat.ge.0 .and. jpat.le.12.and. flag0 .and. ccdlimsel) then
      ypos = ypos + rawyCol(i-1)
      npos = npos+1
    end if
    if (hasrawx .and. jpat.ge.0 .and. jpat.le.12.and. flag0 .and. ccdlimsel) then
      xavpos = xavpos + rawxCol(i-1)
      nxpos = nxpos+1
    end if
!
! accumulate pattern types and energies of pattern types in bands
!
    if (kpat .le. maxpat_plus_1) then
      do j = 0,nband
        if (piCol(i-1).ge.e_low(j).and.piCol(i-1).lt.e_hig(j)) then
          e_sum(j,kpat) = e_sum(j,kpat) + piCol(i-1)
          kband(j,kpat) = kband(j,kpat) + 1.0
        endif
      enddo
    endif
!
! find the energy bin corresponding to event PI channel
!
    kbin = piCol(i-1)/econv

    if (kpat.ge.1 .and. kpat.le.maxpat_plus_1 .and.   &
        kbin.ge.1 .and. kbin.le.maxbin) then
      khis(kpat,kbin) = khis(kpat,kbin) + 1
    endif
!
! doubles HORI + VERT in addition separately
!
    if (kkpat .gt. 0 .and.   &
        kbin.ge.1 .and. kbin.le.maxbin) then
      khis(kkpat,kbin) = khis(kkpat,kbin) + 1
    endif
  enddo ! loop over all events

!
! DETX and DETY average for EPIC-pn
!
  if (is_epn .and. hasdetx .and. hasdety) then
    nsel = COUNT(MASK = (rawxCol(0:ntrow-1).ge.rawxmin .AND. &
                         rawxCol(0:ntrow-1).le.rawxmax .AND. &
                         rawyCol(0:ntrow-1).ge.rawymin .AND. &
                         rawyCol(0:ntrow-1).le.rawymax .AND. &
                         ccdnrCol(0:ntrow-1).ge.ccdmin .AND. &
                         ccdnrCol(0:ntrow-1).le.ccdmax) .AND. &
                       ((flagCol(0:ntrow-1) == 0) .OR. .NOT. (withflag)))
    if (nsel .gt. 0) then
      xdavpos = SUM(1.0*detxCol(0:ntrow-1), &
                 MASK = (rawxCol(0:ntrow-1).ge.rawxmin .AND. &
                         rawxCol(0:ntrow-1).le.rawxmax .AND. &
                         rawyCol(0:ntrow-1).ge.rawymin .AND. &
                         rawyCol(0:ntrow-1).le.rawymax .AND. &
                         ccdnrCol(0:ntrow-1).ge.ccdmin .AND. &
                         ccdnrCol(0:ntrow-1).le.ccdmax) .AND. &
                       ((flagCol(0:ntrow-1) == 0) .OR. .NOT. (withflag)))/nsel
      ydavpos = SUM(1.0*detyCol(0:ntrow-1), &
                 MASK = (rawxCol(0:ntrow-1).ge.rawxmin .AND. &
                         rawxCol(0:ntrow-1).le.rawxmax .AND. &
                         rawyCol(0:ntrow-1).ge.rawymin .AND. &
                         rawyCol(0:ntrow-1).le.rawymax .AND. &
                         ccdnrCol(0:ntrow-1).ge.ccdmin .AND. &
                         ccdnrCol(0:ntrow-1).le.ccdmax) .AND. &
                       ((flagCol(0:ntrow-1) == 0) .OR. .NOT. (withflag)))/nsel
      write(detxtxt,'(a,sp,f8.1,ss)') ' DETX = ',xdavpos
      write(detytxt,'(a,sp,f8.1,ss)') ' DETY = ',ydavpos
    else
      detxtxt = '                                        '
      detytxt = '                                        '
    endif
  else
    withdetxy = .FALSE.
    detxtxt = '                                        '
    detytxt = '                                        '
  endif

  write(errstr,'(1x,a,t35,12l1)') 'CCDs in events file: ', &
                (hasCCD(iccd,node),iccd=1,nccd)
  call message(AppMsg,Verbose)

  write(errstr,'(1x,a,t35,i9  )') 'Number of recognized photons: ',npos
  call message(AppMsg,Verbose)
  write(errstr,'(1x,a,t35,i9  )') 'Number of unrecognized cases: ',mpos
  call message(AppMsg,Verbose)

! +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

  if (withbackgroundset) then
    khis=max(0,khis-nint(real(khis_bkg)/backscal))
    kband=max(0.,kband-kband_bkg/backscal)
!    e_sum=max(0.,e_sum-e_sum_bkg/backscal)
  end if
!
! compute the total intensity (i.e. all valid patterns) per energy band
!   and store this quantity into index "maxpat_plus_2"
!
  do j = 0,nband
    do kpat = 1,maxpat
      e_sum(j,maxpat_plus_2) = e_sum(j,maxpat_plus_2)+e_sum(j,kpat)
      kband(j,maxpat_plus_2) = kband(j,maxpat_plus_2)+kband(j,kpat)
    enddo
  enddo

! single and double fraction with error for band 0
  do kpat = 1,maxpat
    if (kband(0,maxpat_plus_2).gt.0) then
      kfrac(kpat) = real(kband(0,kpat))/real(kband(0,maxpat_plus_2))
      kerror(kpat)= sqrt(kband(0,kpat)/kband(0,maxpat_plus_2)**2 * &
                      (1.+kband(0,kpat)/kband(0,maxpat_plus_2)))
!     write(errstr,*) 'KPAT,KFRAC,KERROR =',KPAT,KFRAC(kpat),KERROR(kpat)
!     call message(AppMsg,Sparse)
    else
      kfrac(kpat) = 0.
      kerror(kpat) = 0.
      nosignal=.true.
    end if
  end do

! do j = 1,nband
!   write(errstr,'(1x,i1,2f8.1,5e11.4)') j,e_low(j),e_hig(j),&
!        (e_sum(j,kpat),kpat=1,maxpat_plus_1)
!   call message(AppMsg,Sparse)
! enddo
  call message(AppMsg,Verbose,' band   E_low   E_hig frac_inv')
  do j = 1,nband
    if (e_sum(j,maxpat_plus_2) .ne. 0) then
      write(errstr,'(4x,i1,2f8.1,f8.5)') j,e_low(j),e_hig(j),&
         e_sum(j,maxpat_plus_1)/e_sum(j,maxpat_plus_2)
    else
      write(errstr,'(4x,i1,2f8.1,f8.5)') j,e_low(j),e_hig(j),0.0
    endif
    call message(AppMsg,Verbose)   
  enddo

  if (is_mos) then
    xpos = xpos_mos
  else
    xpos = xpos_epn
  endif
  if (npos > 0) then
    ypos = ypos/npos
  else
    if (is_mos) then
      ypos = ypos_mos
    else
      ypos = ypos_epn
    endif
  endif
  if (nxpos > 0) then
    xavpos = xavpos/nxpos
  endif

  if ((calMode == 4 .or. calMode == 5).and.(is_EPN)) then
    if (hasAttribute(events,"SRCPOS")) then
      srcpos = int16Attribute(events,"SRCPOS")
    else
      srcpos = 190
      call message(AppMsg,Sparse,'SRCPOS attribute not found')
    endif
    ypos = srcpos
  endif

  if (is_EPN) then
    if (hasAttribute(events,"SRCPOSX")) then
      srcposx = real32Attribute(events,"SRCPOSX")
      write(srcxtxt,'(a,f7.2)') ' SRCX = ',srcposx
    else
      srcposx = -1
      call message(AppMsg,Sparse,'SRCPOSX attribute not found')
      srcxtxt  = '                                        '
    endif
    if (hasAttribute(events,"SRCPOSY")) then
      srcposy = real32Attribute(events,"SRCPOSY")
      write(srcytxt,'(a,f7.2)') ' SRCY = ',srcposy
    else
      srcposy = -1
      call message(AppMsg,Sparse,'SRCPOSY attribute not found')
      srcytxt  = '                                        '
    endif
  endif



  write(errstr,'(1x,a,t35,f9.3,t50,a)') &
       'Average RAWY position: ',ypos,trim(detytxt)
  call message(AppMsg,Verbose)
  write(errstr,'(1x,a,t35,f9.3,t50,a)') &
       'Average RAWX position: ',xavpos,trim(detxtxt)
  if (nxpos > 0) call message(AppMsg,Verbose)

  if (enable_Y) then
    write(errstr,'(1x,a,t35,f9.3)') 'User-defined RAWY position: ',userY
    call message(AppMsg,Verbose)
    ypos = userY
  endif

  if (ypos < 10) then
    write(rawytxt,'(a,f3.1)') 'RAWY = ',ypos
  elseif (ypos < 100) then
    write(rawytxt,'(a,f4.1)') 'RAWY = ',ypos
  else
    write(rawytxt,'(a,f5.1)') 'RAWY = ',ypos
  endif
  if (xavpos < 10) then
    write(rawxtxt,'(a,f3.1)') 'RAWX = ',xavpos
  elseif (xavpos < 100) then
    write(rawxtxt,'(a,f4.1)') 'RAWX = ',xavpos
  else
    write(rawxtxt,'(a,f5.1)') 'RAWX = ',xavpos
  endif

  if (withflag) call release(flag)
  if (withccdnr) call release(ccdnr)
  call release(pattern)
  call release(pi)
  if (hasrawx) call release(rawx)
  if (hasrawy) call release(rawy)
  if (hasdetx) call release(detx)
  if (hasdety) call release(dety)
!
! output mask stuff (FLAG==0)
!
  if (withoutputmask) then

    node = 0
    nhas = 0
    mhas = 0
    CALstat_node = node
    EventsFileType = 0

    hasBADPIX(0:nccd,0:1) = .false.
    hasOFFSET(0:nccd,0:1) = .false.

    do iccd = 0,nccd
      if (iccd == 0) then
        write(blockname(iccd), '(a8)') 'BADPIX  '
        write(blockname2(iccd),'(a8)') 'OFFSETCO'
      else
        write(blockname(iccd), '(a6,i2.2)') 'BADPIX',iccd
        write(blockname2(iccd),'(a6,i2.2)') 'OFFSET',iccd
      endif

      if (hasBlock(infile,blockname(iccd))) then
        hasBADPIX(iccd,node) = .true.
        if (iccd>0) nhas = nhas+1
      endif
      if (hasBlock(infile,blockname2(iccd))) then
        hasOFFSET(iccd,node) = .true.
        if (iccd>0) mhas = mhas+1
      endif
    enddo

    if (hasBADPIX(0,0)) then
      singleChip = .true.
      nhas = 1
    elseif (nhas > 0) then
      singleChip = .false.
    else
      singleChip = .false.
      badpix_from_events = .false.
    endif

    if (singleChip) then
      if (hasBlock(infile,'OFFSETCO')) then      ! old-style single chip file
        EventsFileType = 1
      elseif (hasBlock(infile,'OFFSETS')) then   ! new-style single chip file
        EventsFileType = 2
        blockname2(0) = 'OFFSETS '               ! -> redefine extension name
        hasOFFSET(0,node) = .true.
        mhas = 1
      else                                       ! very old single chip file
        EventsFileType = 6                       ! -> not supported
        offsets_from_events = .false.
      endif
    else
      if (mhas > 0) then                         ! old-style epchain event file
        EventsFileType = 3
      elseif (hasBlock(infile,'OFFSETS')) then   ! new-style epchain event file
        EventsFileType = 4
        blockname2(1:nccd) = 'OFFSETS '          ! -> redefine extension names
        hasOFFSET(1:nccd,node) = .true.
        mhas = 1
      else                                       ! old-style LUX/PPS event file
        EventsFileType = 5                       ! -> not supported
        offsets_from_events = .false.
      endif
    endif
    write(errstr,'(1x,a,t35,i9,2x,a,l1,a)') 'Event file type (single): ', &
         EventsFileType,' (',singleChip,')'
    call message(AppMsg,Noisy)
    write(errstr,'(1x,a,t35,i9,i5)') 'Extensions for BADPIX and OFFSET: ', &
         nhas,mhas
    call message(AppMsg,Noisy)

    if (nhas == 0) then    ! no BADPIX extension
      call warning("noBadpix","No BADPIX extension in input data set")
      badpix_from_events = .false.
      call message(AppMsg,Verbose,' Output mask will use BADPIX from CCF instead')
      call message(AppMsg,Verbose,' This is not yet implemented ! No action performed.')
    else
      badpix_from_events = .true.
      call message(AppMsg,Verbose,' Output mask will use BADPIX from events file')
    endif

    if (badpix_from_events) then
      do iccd = 0, nccd
        map_array(1:ndx,1:ndy) = 1
        CALstat_ccd = iccd
        nsrow = 0
        if (hasBADPIX(iccd,node)) then
          badpix = table(infile,blockname(iccd),AccessType)
          if (hasAttribute(badpix,"WINDOWX0")) then
            xw1 = int16Attribute(badpix,"WINDOWX0")
          else
            call message(AppMsg,Verbose,'WINDOWX0 attribute not found')
          endif
          if (hasAttribute(badpix,"WINDOWDX")) then
            nx = int16Attribute(badpix,"WINDOWDX")
          else
            call message(AppMsg,Verbose,'WINDOWDX attribute not found')
          endif
          xw2 = xw1+nx-1

          if (hasAttribute(badpix,"WINDOWY0")) then
            yw1 = int16Attribute(badpix,"WINDOWY0")
          else
            call message(AppMsg,Verbose,'WINDOWY0 attribute not found')
          endif
          if (hasAttribute(badpix,"WINDOWDY")) then
            ny = int16Attribute(badpix,"WINDOWDY")
          else
            call message(AppMsg,Verbose,'WINDOWDY attribute not found')
          endif
          yw2 = yw1+ny-1
          write(errstr,'(1x,a,t35,i9,5(2x,i5))') 'Window parameters: ', &
                xw1,nx,xw2,yw1,ny,yw2
          call message(AppMsg,Noisy)

          nsrow = numberOfRows(badpix)
          write(errstr,'(1x,a,t35,i9)') 'Number of bad pixel entries: ', nsrow
          call message(AppMsg,Noisy)
          if (nsrow > 0) then
            rawx = column(badpix,"RAWX",AccessType)
            rawy = column(badpix,"RAWY",AccessType)
            yextent = column(badpix,"YEXTENT",AccessType)
            rawxCol => int16Data(rawx)
            rawyCol => int16Data(rawy)
            yextentCol => Int16Data(yextent)
          endif
          if (singleChip .and. iccd == 0) then
            if (is_epn) then
              quadrant = int8attribute(badpix,"QUADRANT")
              ccdid = int8attribute(badpix,"CCDID")
              CALstat_ccd = 3*quadrant+ccdid+1        
              call CAL_setState(CALstat_inst,              &
                    CALstat_ccd, ccdModeId = CALstat_mode, &   ! CALstat_node,
                    accuracyLevel = ACCURACY_HIGH)
            elseif (is_mos) then
              ccdid = int8attribute(badpix,"CCDID")
              CALstat_ccd = ccdid        
              call CAL_setState(CALstat_inst,              &
                    CALstat_ccd, CALstat_node, ccdModeId = CALstat_mode, &
                    accuracyLevel = ACCURACY_HIGH)
            endif
          elseif (.not.singleChip .and. iccd > 0) then
            call CAL_setState(CALstat_inst,    &
                 CALstat_ccd, ccdModeId = CALstat_mode, &   ! **HB**
!                 CALstat_ccd, ccdModeId = CALstat_mode, &   ! CALstat_node,
                 accuracyLevel = ACCURACY_HIGH)
          else
            call message(AppMsg,Sparse,' ... BADPIX and BADPIXnn extension ...')
          endif

          call message(AppMsg,Noisy,' ... FLAG mask: BADPIX ...')
          do irow = 0,nsrow-1
            iex = yextentCol(irow)
            ix0 = rawxCol(irow)
            iy0 = rawyCol(irow)
            iy1 = min(iy0+iex-1,ndy)
            map_array(ix0,iy0) = 0
            do iy = iy0,iy1
              map_array(ix0,iy) = 0
              if (ix0.gt.1) map_array(ix0-1,iy) = 0
              if (ix0.lt.ndx) map_array(ix0+1,iy) = 0
            enddo
            if (iy0.gt.1) map_array(ix0,iy0-1) = 0
            if (iy1.lt.ndy) map_array(ix0,iy1+1) = 0
          enddo

          if (nsrow > 0) then
            call release(rawx)
            call release(rawy)
            call release(yextent)
          endif
!
! set FOV map
!
          if ((calMode == 4 .or. calMode == 5).and.(is_EPN)) then
            call message(AppMsg,Noisy,' ... FLAG mask: FOVmap ... not done for TI+BU mode')
          else
            call message(AppMsg,Noisy,' ... FLAG mask: FOVmap ...')
            call CAL_getFOVmap(FOVmap)
            do iy0 = 1,ndy
              iy = iy0-1
              do ix0 = 1,ndx
                ix = ix0-1
                map_array(ix0,iy0) = map_array(ix0,iy0)*FOVmap(ix,iy)
              enddo
            enddo
          endif
     !
     ! set read-out window map
     !
          call message(AppMsg,Noisy,' ... FLAG mask: WINDOW ...')
          if (xw1 .ge. 1 .and. xw2 .ge. xw1 .and. &
            yw1 .ge. 1 .and. yw2 .ge. yw1 ) then
              map_array(1:xw1,  1:ndy)   = 0
            map_array(xw2:ndx,1:ndy)   = 0
            map_array(1:ndx,  1:yw1)   = 0
            map_array(1:ndx,  yw2:ndy) = 0
          else
             write(errstr,*) 'Window parameters: XW1,NX,XW2,YW1,NY,YW2 = ', &
                              xw1,nx,xw2,yw1,ny,yw2
            call warning('invalidWindow')
          endif
     !
     ! set offset column map
     !
          call message(AppMsg,Noisy,' ... FLAG mask: OFFSET ...')
          if (hasOFFSET(iccd,node)) then
            offset = table(infile,blockname2(iccd),AccessType)
            nsrow = numberOfRows(offset)
            write(errstr,'(1x,a,t35,i9)') 'Number of offset columns: ', nsrow
            call message(AppMsg,Noisy)
            if (nsrow > 0) then
              rawx = column(offset,"RAWX",AccessType)
              rawxCol => int16Data(rawx)
              ccdnr = column(offset,'CCDNR',READ)
              ccdnrCol => int8data(ccdnr)

              do irow = 0,nsrow-1
                if (ccdnrCol(irow) == CALstat_ccd) then
                  ix = rawxCol(irow)
                  map_array(ix,1:ndy) = 0
                  if (ix.gt.1)   map_array(ix-1,1:ndy) = 0
                  if (ix.lt.ndx) map_array(ix+1,1:ndy) = 0
                endif
              enddo

              call release(rawx)     ! column
              call release(ccdnr)    ! column
            endif
            call release(offset)   ! extension
          else
            write(errstr,*) 'Does not have OFFSET for CCD = ',iccd,' [',&
                      trim(blockname2(iccd)),']'
            call message(AppMsg,Noisy)
          endif

          write(outmaskname,'(a,i2.2,a)') trim(omsk1),CALstat_ccd,trim(omsk2)
          write(errstr,*) '... FLAG mask: ',trim(outmaskname)
          call message(AppMsg,Verbose)
          outmask = dataSet(outmaskname,Create)
          map = addArray(outmask, "PRIMARY", Integer8, imageDim)
          map_arrayP => int8Array2Data(map)
          map_arrayP = map_array
          call setAttributes(attributable(map),attributable(badpix))
          if (hasAttribute(map,"HISTORY")) then
            call deleteAttribute(map,"HISTORY")
          else
            call message(AppMsg,Noisy,' ... No HISTORY keyword.')
          endif
          if (hasAttribute(map,"COMMENT")) then
            call deleteAttribute(map,"COMMENT")
          else
            call message(AppMsg,Noisy,' ... No COMMENT keyword.')
          endif
          call release(map)
          call release(outmask)
          call release(badpix)
        endif ! hasBADPIX
      enddo   ! loop over iccd=0,nccd
    endif     ! badpix_from_events
    if (.not. offsets_from_events) then
      call warning('noOFFSET','Cannot reliably handle OFFSETS, not included in FLAG map.')
    endif
  endif       ! outputmask requested

  if (is_epn) then
!   call CAL_setState(infile)
    if (hasAttribute(events,"CCDID") .and. &
        hasAttribute(events,"QUADRANT")) then
      quadrant = int8attribute(events,"QUADRANT")
      ccdid = int8attribute(events,"CCDID")
      CALstat_ccd = 3*quadrant+ccdid+1
    else
      CALstat_ccd = 4             
    end if
    call CAL_setState(CALstat_inst,             &
         CALstat_ccd, ccdModeId = CALstat_mode, &   ! CALstat_node,
         accuracyLevel = ACCURACY_HIGH)
  end if

  if (allocated(map_array)) deallocate(map_array)
!
!-----------------------------------------------------------------------
!
!     accumulate the distributions
!
  call message(AppMsg,Sparse,' ')
  call message(AppMsg,Verbose,' Pattern statistics for recognized events:')

  ks = 0
  kd = 0
  kt = 0
  kq = 0
  kv = 0 ! vertical doubles
  kh = 0 ! horizontal doubles
  ymax = 0.0
  ymin = 1.0e+9

!
  do ibin = 1, maxbin
    ks = ks + khis(1,ibin)
    kd = kd + khis(2,ibin)
    kt = kt + khis(3,ibin)
    kq = kq + khis(4,ibin)
    kv = kv + khis(7,ibin) ! vertical doubles
    kh = kh + khis(8,ibin) ! horizontal doubles
    khis(maxpat_plus_2,ibin) = &
                khis(1,ibin) + khis(2,ibin) + khis(3,ibin) + khis(4,ibin)
  enddo
!
  k = ks + kd + kt + kq
!
  if (k>0) then
    hs = real(ks) / real(k)
    hd = real(kd) / real(k)
    ht = real(kt) / real(k)
    hq = real(kq) / real(k)
    hv = real(kv) / real(k) ! vertical doubles
    hh = real(kh) / real(k) ! horizontal doubles
  else
    call error("noSignal", &
               "No net signal in input set. Wrong backscal parameter ?")
  end if

  write(errstr,'(4(7x,a1))') 's','d','t','q'
  call message(AppMsg,Sparse)
  write(errstr,'(4f8.4)') hs,hd,ht,hq
  call message(AppMsg,Sparse)
!
! compute the pattern distributions vs. channel:
!
  call message(AppMsg,Noisy,' ... compute pattern distributions ...')
  do ipat = 1, maxpat_plus_4
    if (ipat .eq. 6) goto 999 ! splitting invalid patterns not yet implemented
    if (ipat .gt. 6 .and. is_mos) goto 999 ! do not plot H/V doubles for MOS
!   print*,'starting IPAT = ',ipat
    h = sigma * sigma
    s = 0.0
    i = 1
    j = 0
    k = 0
    do while (khis(ipat,i).eq.0 .and. i.lt.maxbin)
!     print*,'IPAT,I,KHIS() =',IPAT,I,KHIS(IPAT,I)
      i = i + 1
    enddo
    jm(1) = i
    if (i.gt.1) i = i - 1
    do while (i.lt.maxbin)
      i = i + 1
      s = s + real(khis(ipat,i))
      if (s*s.gt.h*s) then
        k = k + 1
        jcts(k) = s
        jp(k) = i
        s = 0.0
      endif
    enddo
    mbin = k
!   print*,'obtained MBIN = ',mbin

    if (mbin > 0) then
      do i = 2, mbin
        jm(i) = jp(i-1) + 1
      enddo

      do i = 1, mbin
        h = real(jp(i)-jm(i)) + 1.0
        sprb(i)  = real(jcts(i)) / h
      enddo
!
! store as histogram
!
      j = j+1
      xhis(ipat,j) = alog10((real(jm(1))-0.5)*conv)
      y = alog10(sprb(1))
      yhis(ipat,j) = y
      j = j+1
      xhis(ipat,j) = alog10((real(jp(1))+0.5)*conv)
      yhis(ipat,j) = y
      ymax = max(ymax,y)
      ymin = min(ymin,y)
      do i = 2, mbin
        j = j+1
        xhis(ipat,j) = xhis(ipat,j-1)
        y = alog10(sprb(i))
        yhis(ipat,j) = y
        j = j+1
        xhis(ipat,j) = alog10((real(jp(i))+0.5)*conv)
        yhis(ipat,j) = y
        ymax = max(ymax,y)
        ymin = min(ymin,y)
      enddo
    else
      write(errstr,'(a,i1)') &
          'not enough statistics for at least 1 bin for PAT = ',ipat
      call warning('sigmaTooLarge')
    endif
    nbin(ipat) = j
 999 continue
  enddo
!
! initialize the plot for pattern distributions:
!
  call message(AppMsg,Noisy,' ')
  call message(AppMsg,Noisy,' ... prepare plot input data set ...')

  if (index(trim(device),"PS") .gt. 0)  is_ps = .true.
  if (index(trim(device),"GIF") .gt. 0) is_gif = .true.

  statusPG = pgopen(trim(device))
  write(errstr,'(a,2x,a,2x,a,i5)') &
          ' PGOPEN for device',device,': STATUS = ',statusPG
  call message(AppMsg,Noisy)

  if (statusPG .gt. 0) then
!   do something
  endif
!
  if (is_ps) then
    call pgbegin(0,trim(outfilename)//trim(device),1,1)
  elseif (is_gif) then
    call pgbegin(0,trim(outfilename)//trim(device),1,1)
  else
    call pgbegin(0,trim(device),1,1)
  endif
  call pgpaper(7.8,1.414)
!
! top bar with some text
!
  vxmin = 0.15 - 0.75*0.12
  vxmax = 0.90
  vymin = 0.98  ! epatplot-1.15: 0.97
  vymax = 1.00  ! epatplot-1.15: 0.99
  grey  = 0.95
!
! call pgscr(255,grey,grey,grey)
  call pgscr(16,0.0,0.0,0.0)    ! black ??
  call pgscr(17,1.0,1.0,1.0)    ! white ??
  call pgscr(18,grey,grey,grey) ! gray ??
  call pgvport(0.,1.,0.,1.)
  call pgsci(17)
  call pgrect(0.0,1.0,0.0,1.0)

  call pgvport(vxmin,vxmax,vymin,vymax)
  call pgwindow(0.0,1.0,0.0,1.0)
  call pgsci(18)
  call pgrect(0.0,1.0,0.0,1.0)
  call pgsci(16)
  call pgsch(0.5)
  call pgslw(1)
  call pgsci(1) ! mjf @ 2011-08-30
  call pgptext(0.99, 0.4, 0.0, 1.0, trim(outfilename))
  call pgptext(0.01, 0.4, 0.0, 0.0, &
           'XMM '//trim(insttxt)// &
           ', pattern statistics: epatplot-'//trim(CodeVersion))
!
! username, date and time at bottom of plot
!
  call pgsch(1.0)
  call pgsci(13) ! mjf @ 2011-08-30
  call pgiden

  call message(AppMsg,Noisy,' ... prepare X and Y ranges ...')
  xmin = alog10(17.0*conv)
  xamin = alog10(17.0)
  xmax = alog10(4200.0*conv)
  xamax = alog10(4200.0)
   
! ymin = max(alog10(1.e-3),real(int(alog10(ymin)-1)))
! ymax = real(int(alog10(ymax)+1))

  if (plotyrange(1) .lt. plotyrange(2)) then
    ymin = plotyrange(1)
    ymax = plotyrange(2)
  else
    ymin = int(max(-3.0,ymin))-1
    ymax = int(ymax)+1
  endif

  write(errstr,*) '... ranges: XMIN,XMAX, YMIN,YMAX = ', &
     xmin,xmax,ymin,ymax
  if (ymin>ymax) then
    call message(AppMsg,Verbose)
    call message(AppMsg,Verbose,' ... YMIN has been set to YMAX-2 ...')
    ymin = ymax - 2
  else
    call message(AppMsg,Noisy+1)
  endif

  xsiz = xmax - xmin
  xasiz = xamax - xamin
  ysiz = ymax - ymin
  vxmin = 0.15
  vxmax = 0.90
  vymin = 0.52
  vymax = 0.90
  call message(AppMsg,Noisy+1,' ... call pgvport(vxmin,vxmax,vymin,vymax) ...')
  call pgvport(vxmin,vxmax,vymin,vymax)
  call message(AppMsg,Noisy+1,' ... call pgwindow(xmin,xmax,ymin,ymax) ...')
  call pgwindow(xmin,xmax,ymin,ymax)
  call message(AppMsg,Noisy+1,' ... call pgsci(18) ...')
  call pgsci(18)
  call message(AppMsg,Noisy+1,' ... call pgrect(xmin,xmax,ymin,ymax) ...')
  call pgrect(xmin,xmax,ymin,ymax)
  call message(AppMsg,Noisy+1,' ... call pgsci(16) ...')
  call pgsci(16)
  call pgslw(3)
  call pgbox('BC',0.0,0,'BC',0.0,0)
  call pgbox('LN',0.0,0,' ',0.0,0)
  call pgsch(0.75)
  call pgbox(' ',0.0,0,'LNV',0.0,0)
  call pgsch(1.0)
  call pgslw(1)
  call pgbox('BCLST',0.0,0,'BCLST',0.0,0)
  call pgslw(5)
  call pgsch(1.1)
  xtxt = (xmax + xmin)*0.5
  ytxt = ymin - ysiz*0.11
  call message(AppMsg,Noisy+1,' ... call pgptext(xtxt, ytxt, 0.0, 0.5, trim(xlabel)) ...')
  call pgptext(xtxt, ytxt, 0.0, 0.5, trim(xlabel))
! call pgsch(1.25)
  call pgsch(1.12)
  call pgslw(4)
  xtxt = xmin - xsiz*0.12
  ytxt = ymax + ysiz*0.12
  call message(AppMsg,Noisy+1,' ... call pgptext(INSTTXT) ...')
  call pgsci(1)
  call pgptext(xtxt, ytxt, 0.0, 0.0, 'XMM / '//trim(insttxt)//' '//trim(instmode)//' '//trim(filter))
  xtxt = xmax
! ytxt = ymax + ysiz*0.12 ! shift it a bit upwards to squeeze in OBJECT:
  ytxt = ymax + ysiz*0.15
  call pgsch(0.65)
  call pgslw(3)
  call pgptext(xtxt, ytxt, 0.0, 1.0, trim(revlab))
! call pgsch(1.0)
  call pgslw(3)
  call pgsch(0.65)
  ytxt = ymax + ysiz*0.05
  xtxt = xmax
  call message(AppMsg,Noisy+1,' ... call pgptext(FILENAME) ...')
  call pgsci(1) ! mjf @ 2011-08-30
  call pgsci(16) ! mjf @ 2011-08-30
  call pgptext(xtxt, ytxt, 0.0, 1.0, trim(infilename))
!
  if (livetime.gt.0) then
    xtxt = xmax - xsiz*0.025
    ytxt = ymax - ysiz*0.060
    call message(AppMsg,Noisy+1,' ... call pgptext(LIVETIME) ...')
    write(errstr,'(1x,a,f8.1,a)') 'Livetime: ',livetime,' [s]'
    call pgsci(13) ! mjf @ 2011-08-30
    call pgptext(xtxt, ytxt, 0.0, 1.0,  trim(errstr))
  endif
!
  ytxt = ymax + ysiz*0.10
  xtxt = xmax
  call message(AppMsg,Noisy+1,' ... call pgptext(OBJECT) ...')
  call pgsci(16)
  call pgptext(xtxt, ytxt, 0.0, 1.0, trim(object))
!
  call pgsch(1.0)
  call pgslw(4)
  xtxt = xmin - xsiz*0.1
! ytxt = ymin + ysiz*0.100
  ytxt = ymin - ysiz*0.20*0.5
  if (is_epn) then
    call pgsch(0.80)
    ipat = 8
    call pgsci(icol(ipat))
    ytxt = ytxt + ysiz*0.15
    call pgptext(xtxt, ytxt, 90.0, 0.5, trim(chpt(ipat))//'/adu')
    ipat = 7
    call pgsci(icol(ipat))
    ytxt = ytxt + ysiz*0.15
    call pgptext(xtxt, ytxt, 90.0, 0.5, trim(chpt(ipat))//'/adu')
    do ipat = maxpat_plus_1,1,-1
      call pgsci(icol(ipat))
      ytxt = ytxt + ysiz*0.15
      call pgptext(xtxt, ytxt, 90.0, 0.5, trim(chpt(ipat))//'/adu')
    enddo
    call pgsch(1.00)
  else
    do ipat = maxpat_plus_1,1,-1
      call pgsci(icol(ipat))
      ytxt = ytxt + ysiz*0.20
      call pgptext(xtxt, ytxt, 90.0, 0.5, trim(chpt(ipat))//'/adu')
    enddo
  endif
  xtxt = xmin + xsiz*0.05
  ytxt = ymin + ysiz*0.05
  call pgsci(icol(maxpat_plus_1))
  write(errstr,'(1x,a,f6.2,a)') 'Energy fraction lost in invalid patterns in [1.0-32.8] keV:',&
       e_sum(nband,maxpat_plus_1)/e_sum(nband,maxpat_plus_2)*100,'%'
  call pgslw(2)
  call pgsch(0.50)
  call pgptext(xtxt, ytxt, 0.0, 0.0, trim(errstr))
  call pgsci(16)
  call pgslw(1)
!
! plot the pattern distributions vs. channel:
!
  call message(AppMsg,Noisy,' ... plot pattern distributions ...')
  do ipat = 1, maxpat_plus_4
    if (ipat .eq. 6) goto 998 ! splitting invalid patterns not yet implemented
    if (ipat .gt. 6 .and. is_mos) goto 998 ! do not plot H/V doubles for MOS
    j = 0
!
! plot as histogram
!
    if (nbin(ipat) .gt. 0) then
      call pgsci(icol(ipat))
      j = j+1
      xdat = xhis(ipat,j)
      ydat = yhis(ipat,j)
      call pgmove(xdat,ydat)
      j = j+1
      xdat = xhis(ipat,j)
      call pgdraw(xdat,ydat)
      do i = j+1, nbin(ipat), 2
        ydat = yhis(ipat,i)
        call pgdraw(xdat,ydat)
        xdat = xhis(ipat,i+1)
        call pgdraw(xdat,ydat)
      enddo
    endif
 998 continue
  enddo
!
! initialize the plot for pattern statistics:
!
  call message(AppMsg,Noisy,' ... plot pattern statistics ...')
  ymin = -0.05
  ymax = 1.05
  ysiz = ymax - ymin
  vymin = vymin - 0.45
  vymax = vymax - 0.45
  call pgvport(vxmin,vxmax,vymin,vymax)
  call pgwindow(xmin,xmax,ymin,ymax)
  call pgsci(18)
  call pgrect(xmin,xmax,ymin,ymax)
  call pgsci(16)
  call pgslw(3)
  call pgbox('BC',0.0,0,'BC',0.0,0)
  call pgbox('LN',0.0,0,' ',0.0,0)
  call pgsch(0.75)
  call pgbox(' ',0.0,0,'NV',0.0,0)
  call pgsch(1.0)
  call pgslw(1)
  call pgbox('A',0.0,0,' ',0.0,0)
  call pgbox('BCLST',0.0,0,'BCST',0.0,0)
! call pgslw(6)
  call pgslw(5)
  call pgsch(1.1)
  xtxt = (xmax + xmin)*0.5
  ytxt = ymin - ysiz*0.11
  call pgptext(xtxt, ytxt, 0.0, 0.5, trim(xlabel))
  call pgslw(3)
  call pgsch(0.7)
!  xtxt = xmin + xsiz*0.95
!  ytxt = ymin + ysiz*0.85

  xtxt = xmin + xsiz*0.025
  ytxt = ymin + ysiz*0.48
  call pgptext(xtxt, ytxt, 0.0, 0.0, trim(RAWYTXT))
  ytxt = ymin + ysiz*0.445
  call pgptext(xtxt, ytxt, 0.0, 0.0, trim(RAWXTXT))
!
  if (withdetxy) then
    xtxt = xtxt + xsiz*0.2
    ytxt = ymin + ysiz*0.48
    call pgptext(xtxt, ytxt, 0.0, 0.0, trim(DETYTXT))
    ytxt = ymin + ysiz*0.445
    call pgptext(xtxt, ytxt, 0.0, 0.0, trim(DETXTXT))
  endif
!
  if (withsrcxy) then
    xtxt = xtxt + xsiz*0.2
    ytxt = ymin + ysiz*0.48
    call pgptext(xtxt, ytxt, 0.0, 0.0, trim(SRCYTXT))
    ytxt = ymin + ysiz*0.445
    call pgptext(xtxt, ytxt, 0.0, 0.0, trim(SRCXTXT))
  endif
!
  xtxt = xmin + xsiz*0.025
  ytxt = ymin + ysiz*0.40
  write(string,"(f3.1,'-',f3.1, ' keV observed-to-model fractions:')") &
                 e_low(0)/1e3, e_hig(0)/1e3
  call pgptext(xtxt, ytxt, 0.0, 0.0, trim(string))
  call pgslw(5)
  call pgsch(1.1)
!  xtxt = xmin + xsiz*0.95
!  ytxt = ymin + ysiz*0.91
  xtxt = xmin + xsiz*0.025
  ytxt = ymin + ysiz*0.52
  if (ccdmode(1:2) .ne. "  ") then
!   if (calMode .ne. 5) then
      call pgptext(xtxt, ytxt, 0.0, 0.0, trim(ccdmode)//' Model')
!   else
!     call pgptext(xtxt, ytxt, 0.0, 1.0, trim(ccdmode)//' (no model yet)')
!   endif
  endif
  call pgslw(4)
  call pgsci(icol(4))
  xtxt = xmin - xsiz*0.1
  ytxt = ymin + ysiz*yt0
  call pgptext(xtxt, ytxt, 90.0, 0.5, 'q')
  call pgsci(icol(3))
  ytxt = ytxt + ysiz*yt1
  call pgptext(xtxt, ytxt, 90.0, 0.5, 't')
  call pgsci(icol(2))
  ytxt = ytxt + ysiz*yt1
  call pgptext(xtxt, ytxt, 90.0, 0.5, 'd')
  call pgsci(icol(1))
  ytxt = ytxt + ysiz*yt1
  call pgptext(xtxt, ytxt, 90.0, 0.5, 's')
  if (is_EPN) then
    call pgsci(6)
    ytxt = ytxt + ysiz*yt1
    call pgptext(xtxt, ytxt, 90.0, 0.5, 's+d')
  endif
  call pgsci(16)
  call pgslw(1)
!
! determine binning for statistics plot
!
  call message(AppMsg,Noisy,' ... determine binning for statistics plot ...')
  ipat = maxpat_plus_2
  h = sigma * sigma * 10.0  * 10.0
  s = 0.0
  i = 1
  k = 0
  do while (khis(ipat,i).eq.0 .and. i.lt.maxbin)
    i = i + 1
  enddo
  jm(1) = i
  if (i.gt.1) i = i - 1
  do while (i.lt.maxbin)
    i = i + 1
    s = s + real(khis(ipat,i))
    if (s*s.gt.h*s) then
      k = k + 1
      jp(k) = i
      s = 0.0
    endif
  enddo
  mbin = k
  do i = 2, mbin
    jm(i) = jp(i-1) + 1
  enddo
!
! determine the fraction of patterns among the recognized photons:
!
  call message(AppMsg,Noisy,&
    ' ... determine the fraction of patterns among the recognized photons ...')
  do ipat = 1, maxpat
    do i = 1, mbin
      k1 = 0
      k2 = 0
      do ibin = jm(i), jp(i)
        k1 = k1 + khis(ipat,ibin)
        k2 = k2 + khis(maxpat_plus_2,ibin)
      enddo
      rhis(ipat,i) = real(k1) / amax1(real(k2),1.0)
    enddo
!
! plot as histogram
!
    call pgsci(icol(ipat))
    xdat = alog10((real(jm(1))-0.5)*conv)
    ydat = rhis(ipat,1)
    call pgmove(xdat,ydat)
    xdat = alog10((real(jp(1))+0.5)*conv)
    call pgdraw(xdat,ydat)
    do i = 2, mbin
      ydat = rhis(ipat,i)
      call pgdraw(xdat,ydat)
      xdat = alog10((real(jp(i))+0.5)*conv)
      call pgdraw(xdat,ydat)
    enddo
  enddo

  if (is_EPN) then
    call pgsci(6)
    xdat = alog10((real(jm(1))-0.5)*conv)
    ydat = rhis(1,1)+rhis(2,1)
    call pgmove(xdat,ydat)
    xdat = alog10((real(jp(1))+0.5)*conv)
    call pgdraw(xdat,ydat)
    do i = 2, mbin
      ydat = rhis(1,i)+rhis(2,i)
      call pgdraw(xdat,ydat)
      xdat = alog10((real(jp(i))+0.5)*conv)
      call pgdraw(xdat,ydat)
    enddo
  endif
!------------------------------------------------------------------------------
! plot the standard single and double pattern statistics
! as used in the response matrix
!
  call message(AppMsg,Noisy,' ... model curve as used for DRM ...')
!
  if (is_mos) then
    CALstat_ccd = 1
    CALstat_node = -1  !AI: variable initialization. it may produce a CAL crash
    call CAL_setState(CALstat_inst,    &
       CALstat_ccd, CALstat_node, ccdModeId = CALstat_mode, &
       accuracyLevel = ACCURACY_HIGH)
  endif
  pfds = CAL_getPatternFractionServer()
!
  mod(1:maxpat) = 0.0
  modnorm(1:maxpat) = 0.0
!  istart=e_low(2)*econv
!  iend=e_hig(2)*econv
  do ipat = 1,2
    model(ipat,1:maxbin) = 0.0
!   call patmod(model(ipat,1:maxbin), maxbin, calMode, dthresh, ipat, ypos)
    iy16 = nint(ypos)
    ix16 = nint(xpos)
    if (.not.special) then
      call CAL_patternFractionAt(pfds, ix16, iy16)
      pattFracChan => CAL_getPatternFractionInChannel(pfds, ipat-1)
      mmp = size(pattFracChan)
      write(errstr,*) 'Number of channels in pattern array: ',mmp
      call message(AppMsg,Verbose)
      model(ipat,1:mmp) = pattFracChan(0:mmp-1)
    else
      calMode = 33
      write(errstr,*) 'Special CalMode assumed for model pattern fractions:',&
                       calMode
      call message(AppMsg,Verbose)
      call patmod(model(ipat,1:maxbin), maxbin, calMode, 207, ipat, 190.0)
      mmp = maxbin
    endif
    if (is_epn) then
      istart=min(mmp,nint(e_low(0)/econv))
      iend=min(mmp,nint(e_hig(0)/econv))
    else
      istart=min(mmp,nint(e_low(0)/chan))
      iend=min(mmp,nint(e_hig(0)/chan))
    end if
    do i=istart,iend
!      mod(ipat)     = mod(ipat)+model(ipat,i)**2*khis(maxpat_plus_2,i)
!      modnorm(ipat) = modnorm(ipat)+model(ipat,i)*khis(maxpat_plus_2,i)
      mod(ipat)=mod(ipat)+model(ipat,i)*khis(maxpat_plus_2,i)
      modnorm(ipat)=modnorm(ipat)+khis(maxpat_plus_2,i)
    end do
    if (modnorm(ipat)>0.) then
      mod(ipat)=mod(ipat)/modnorm(ipat)
    else
      mod(ipat)=0.
    end if
    if (mod(ipat)>0.) then
      kfrac(ipat)=kfrac(ipat)/mod(ipat)
      kerror(ipat)=kerror(ipat)/mod(ipat)
    else
      kfrac(ipat)=0.
      kerror(ipat)=0.
      nosignal=.true.
    end if
!   write(errstr,*) 'ISTART,IEND = ',istart,iend
!   call message(AppMsg,Sparse)
!   write(errstr,*) "==>",ipat,kfrac(ipat),kerror(ipat),mod(ipat),modnorm(ipat)
!   call message(AppMsg,Sparse)
!    print *,"==>",ipat,kfrac(ipat),kerror(ipat),mod(ipat)
  enddo

  call message(AppMsg,Sparse," ")
  if (nosignal) call warning("noSignal", &
        "No net signal in observed-to-model band. Wrong energy range ?") 
  write(errstr,"(f3.1,'-',f3.1, ' keV observed-to-model fractions:')") &
                 e_low(0)/1e3, e_hig(0)/1e3
  call message(AppMsg,Sparse)
!  call message(AppMsg,Sparse,"0.5-2.0 keV observed-to-model fractions:")
  write(errstr,'(2(a,f6.3,a,f6.3))') "s:",kfrac(1)," +/-",kerror(1), &
                                  "   d:",kfrac(2)," +/-",kerror(2)
  call message(AppMsg,Sparse)
  call message(AppMsg,Sparse," ")
  sdfractxt=errstr
  call pgsci(16)
  call pgslw(3)
  call pgsch(0.7)
  xtxt = xmin + xsiz*0.025
  ytxt = ymin + ysiz*0.36
  call pgptext(xtxt, ytxt, 0.0, 0.0, sdfractxt(1:18))
  ytxt = ymin + ysiz*0.32
  call pgptext(xtxt, ytxt, 0.0, 0.0, sdfractxt(22:39))
  call pgslw(1)

  if (.not. compressed) then
    write(string,"(f3.1,'-',f3.1, ' keV')") e_low(0)/1e3, e_hig(0)/1e3
    call setAttribute(events,"SNGL_OTM",kfrac(1),"", &
                      trim(string)//" observed-to-model singles fraction")
    call setAttribute(events,"ESGL_OTM",kerror(1),"", &
                      "obs-to-mod singles fraction error")
    call setAttribute(events,"DBLE_OTM",kfrac(2),"", &
                      trim(string)//" observed-to-model doubles fraction")
    call setAttribute(events,"EDBL_OTM",kerror(2),"", &
                      "obs-to-mod doubles fraction error")
  end if
!
! plot as line
!
  call message(AppMsg,Noisy,' ... plot as line ...')
  if (is_epn) then
    call pgwindow(xamin,xamax,ymin,ymax)
    do ipat = 1,2
      call pgsci(icol(ipat))
      i = 20
      xdat = alog10(real(i))
      ydat = model(ipat,i)
      call pgmove(xdat,ydat)
      do i = 21, mmp
        xdat = alog10(real(i))
        ydat = model(ipat,i)
        call pgdraw(xdat,ydat)
      enddo
    enddo
! s+d:
    call pgsci(6)
    i = 20
    xdat = alog10(real(i))
    ydat = model(1,i)+model(2,i)
    call pgmove(xdat,ydat)
    do i = 21, mmp
      xdat = alog10(real(i))
      ydat = min(model(1,i)+model(2,i),1.0)
      call pgdraw(xdat,ydat)
    enddo
  else
    do ipat = 1,2
      call pgsci(icol(ipat))
      i = 2
!     call message(AppMsg,Noisy,' ... !EPN: starting point ...')
      xdat = alog10(real(i)*chan/rconv)
      ydat = model(ipat,i)
      call pgmove(xdat,ydat)
      do i = 3, mmp
          xdat = alog10(real(i)*chan/rconv)
          ydat = model(ipat,i)
          call pgdraw(xdat,ydat)
      enddo
!     call message(AppMsg,Noisy,' ... !EPN: loop done...')
    enddo
  endif
  call message(AppMsg,Noisy,' ... call pgend ...')
! call pgend
  call pgclos

  call release(events)
  call release(infile)
!
! write out results into ASCII file which can be plotted with QDP
!
  if (enable_qdp) then
    call get_io_Unit(iw)
    open(iw,file=outqdpname,status='unknown',iostat=ierr)
    if (ierr .eq. 0) then
      write(iw,'(a,a)') '! ',trim(outqdpname)
      write(iw,'(a)')'read serr 1'
      write(iw,'(a)')'col 0 on 2 3'
      write(iw,'(a)')'error off 8 9 10 11'
      write(iw,'(a)')'line on 8 9 10 11'
      write(iw,'(a)')'label x ADU Channl'
      write(iw,'(a)')'label y Pattern fraction'
      write(iw,'(a)')'resc y -0.05 1.05'
      write(iw,'(a)')'log x'
      write(iw,'(a)') '! CH           CHmin  CHmax  s'//&
                      '        d        t        q        s_cali   d_cali   t_cali   q_cali'
      write(iw,'(a)') '! -----        -----  -----  ------   '//&
                      '------   ------   ------   ------   ------   ------   ------'
  !
      do i = 1, mbin
        h = (jp(i)+jm(i))*0.5
        he = (jp(i)-jm(i)+1)*0.5
        mod(1:maxpat)=0.0
        do ipat=1,4
          do j=jm(i),jp(i)
            if (is_mos) then
              mod(ipat)=mod(ipat)+model(ipat,nint(j*econv/chan))
            else
              mod(ipat)=mod(ipat)+model(ipat,j)
            end if
          end do
          mod(ipat)=mod(ipat)/real(jp(i)-jm(i)+1)
        end do    

        write(iw,'(2f6.1,2i7,9f9.4)') h,he,jm(i),jp(i),&
            rhis(1,i),rhis(2,i),rhis(3,i),rhis(4,i), &
            mod(1),mod(2),mod(3),mod(4) ! ,rhis(5,i)
      enddo
      close(iw)
!     call message(AppMsg,Sparse,' ')
      write(errstr,*) 'QDP output written to file ' // trim(outqdpname)
    else
      write(errstr,*) 'QDP output file could not be opened:' // trim(outqdpname)
    endif
    call message(AppMsg,Verbose)
    call message(AppMsg,Verbose,' ')
! else
!   call message(AppMsg,Verbose,' ')
  endif

  write(errstr,*) 'PostScript output written to file ' // trim(outfilename)
  call message(AppMsg,Sparse)

 990 continue ! empty events extension

  call message(AppMsg,Verbose,' ')

end subroutine epatplot


! =========================
! Program parameter reading
! =========================
subroutine getInParamValues(infilename, sigma, device, outdir, withflag, &
                            useplotfile, plotfile, xaxisadu, enable_qdp, &
                            enable_pum, userMode, enable_Y, userY,       &
                            withoutputmask, outmaskname, ccdlimits,      &
                            nccdlimits, plotyrange, withbackgroundset,   &
                            bkgfilename, backscal, modifyInSet, pileuprange, &
                            withdetxy,withsrcxy)

  integer            ,intent(in)    :: nccdlimits
  integer            ,intent(out)   :: userMode,ccdlimits(1:nccdlimits)
  real(kind=single)  ,intent(out)   :: sigma, userY, backscal, &
                                       plotyrange(1:2), pileuprange(2)
  logical(kind=bool) ,intent(out)   :: useplotfile,xaxisadu,enable_qdp,&
                                       withflag, enable_pum, enable_Y, &
                                       withoutputmask,withbackgroundset, &
                                       modifyInSet, withdetxy, withsrcxy
  character(len=*)   ,intent(out)   :: infilename,device,outdir,plotfile,&
                                       outmaskname,bkgfilename
  integer                           :: i,npar

  infilename  = trim(stringParameter('set'))
  sigma       = realParameter('sigma')
  userMode    = intParameter('usermode')
  enable_pum  = booleanParameter('withusermode')
  userY       = realParameter('userrawy')
  enable_Y    = booleanParameter('withuserrawy')
  withflag    = booleanParameter('withflag')
  withoutputmask = booleanParameter('withoutputmask')
  outmaskname = stringParameter('outmaskname')
  useplotfile = booleanParameter('useplotfile')
  device      = trim(stringParameter('device'))
  outdir      = trim(stringParameter('outdir'))
  plotfile    = trim(stringParameter('plotfile'))
  xaxisadu    = booleanParameter('xaxisadu')
  enable_qdp  = booleanParameter('withqdp')
  withbackgroundset = booleanParameter('withbackgroundset')
  bkgfilename = trim(stringParameter('backgroundset'))
  backscal = realParameter('backscal')
  modifyInSet = booleanParameter('modifyinset')
  withdetxy   = booleanParameter('withdetxy')
  withsrcxy   = booleanParameter('withsrcxy')

  if (parameterCount('pileupnumberenergyrange') .ne. 2) &
    call error ("wrongParam","wrong pileupnumberenergyrange")
  pileuprange(1) = realParameter('pileupnumberenergyrange',0)
  pileuprange(2) = realParameter('pileupnumberenergyrange',1)
  if (pileuprange(1) >= pileuprange(2)) &
    call error ("wrongParam","wrong pileupnumberenergyrange")

  npar        = parameterCount('ccdlimits')
  do i = 0, npar-1
    ccdlimits(i+1) = intParameter('ccdlimits',i)
  enddo

  npar        = parameterCount('plotyrange')
  do i = 0, npar-1
    plotyrange(i+1) = realParameter('plotyrange',i)
  enddo

  if (verbosity(AppMsg) > Noisy) then
    call message(AppMsg,Sparse,'List of parameters as read:')
    write(errstr,*) 'set             -> infilename       ',infilename
    call message(AppMsg,Sparse)
    write(errstr,*) 'sigma           -> sigma            ',sigma
    call message(AppMsg,Sparse)
    write(errstr,*) 'usermode        -> userMode         ',userMode
    call message(AppMsg,Sparse)
    write(errstr,*) 'withusermode    -> enable_pum       ',enable_pum
    call message(AppMsg,Sparse)
    write(errstr,*) 'userrawy        -> userY            ',userY
    call message(AppMsg,Sparse)
    write(errstr,*) 'withuserrawy    -> enable_Y         ',enable_Y
    call message(AppMsg,Sparse)
    write(errstr,*) 'withflag        -> withflag         ',withflag
    call message(AppMsg,Sparse)
    write(errstr,*) 'withoutputmask  -> withoutputmask   ',withoutputmask
    call message(AppMsg,Sparse)
    write(errstr,*) 'outmaskname     -> outmaskname      ',outmaskname
    call message(AppMsg,Sparse)
    write(errstr,*) 'useplotfile     -> useplotfile      ',useplotfile
    call message(AppMsg,Sparse)
    write(errstr,*) 'device          -> device           ',device
    call message(AppMsg,Sparse)
    write(errstr,*) 'outdir          -> outdir           ',outdir
    call message(AppMsg,Sparse)
    write(errstr,*) 'plotfile        -> plotfile         ',plotfile
    call message(AppMsg,Sparse)
    write(errstr,*) 'xaxisadu        -> xaxisadu         ',xaxisadu
    call message(AppMsg,Sparse)
    write(errstr,*) 'withqdp         -> enable_qdp       ',enable_qdp
    call message(AppMsg,Sparse)
    write(errstr,*) 'modifyinset     -> modifyInSet      ',modifyInSet
    call message(AppMsg,Sparse)
    write(errstr,*) 'withdetxy       -> withdetxy        ',withdetxy
    call message(AppMsg,Sparse)
    write(errstr,*) 'withsrcxy       -> withsrcxy        ',withsrcxy
    call message(AppMsg,Sparse)
  endif

! if (.NOT. xaxisadu) then
!   call message(AppMsg,Sparse,'Parameter setting XAXISADU=N not ' // &
!               'implemented yet, using XAXISADU=Y in the following.')
!   xaxisadu = .true.
! endif

end subroutine getInParamValues


! ==========================================
! Create output filename from input filename
! ==========================================
subroutine getCanonicalName(infilename,canonicalname)

  implicit none
  character(len=*)   ,intent(in)    :: infilename
  character(len=*)   ,intent(out)   :: canonicalname
  integer                           :: ns,nd,nl,il

  nl = len(trim(infilename))
  if (index(infilename,".gz")>0) nl=nl-3
  do il = nl, 1, -1
    if (infilename(il:il) == '/') exit
  enddo
  ns = il+1 ! max(il,1)
! print*,'IL,NS,NL = ',il,ns,nl
  do il = nl, ns, -1
    if (infilename(il:il) == '.') exit
  enddo
  if (il <= ns) then
    nd = nl
  else
    nd = max(il,2)-1
  endif
! print*,'IL,ND = ',il,nD
  canonicalname = infilename(ns:nd) // '_pat'
! print*,'|',trim(canonicalname),'|'

end subroutine getCanonicalName


! ==========================================
! Create output filename from input filename
! ==========================================
subroutine stripSlash(infilename)

  implicit none
  character(len=*)   ,intent(inout) :: infilename
  integer                           :: nl,il,jl

  nl = len(trim(infilename))
  do il = nl, 1, -1
    if (infilename(il:il) == '/') then
      do jl = il-1, 1, -1
        if (infilename(jl:jl) == '/') then
          infilename(jl:nl) = infilename(jl+1:nl)//' '
        else
          exit
        endif
      enddo
    endif
  enddo
  nl = len(trim(infilename))
  do il = nl, 3, -1
    if (infilename(il-2:il) == '/./') then
      infilename(il-2:nl) = infilename(il:nl)//'  '
    endif
  enddo
  if (infilename(1:2) == './') then
    infilename(1:nl) = infilename(3:nl)//'  '
  endif
! print*,'|',trim(infilename),'|'

end subroutine stripSlash


! ==========================================================================
! suggested by Clive Page in SSC-LUX-TN-0005 (Issue 1.0, 1996 Sept 2)
! ==========================================================================
subroutine get_io_unit(iounit)

  implicit none
  integer, intent(out) :: iounit
  logical              :: lopen, lexist
  do iounit = 99,0,-1
    inquire(unit=iounit, exist=lexist, opened=lopen)
    if (lexist .and. .not. lopen) exit
  enddo

end subroutine get_io_unit


! ==========================================================================
! returns model pattern fractions for SW mode with 100 adu lower threshold
! derived empirically from merged MCG--6-30-15 data
! to be used for 0056_01223307010_PNU002 (Crab)
! ==========================================================================
subroutine get_mjf_model()

!  implicit none

end subroutine get_mjf_model


! ==========================================================================
! returns the pattern distribution model
! ==========================================================================
subroutine patmod(patfraction, maxbin, calMode, dthresh, pattyp, line)

  implicit none
  integer          ,intent(in)    :: calMode, dthresh, pattyp, maxbin
  real(kind=single),intent(in)    :: line
  real(kind=single),intent(out)   :: patfraction(1:maxbin)
  integer                         :: ibin
  real(kind=single)               :: rbin, rbinscal

  integer                         :: ndthresh, irun = 0
  integer                            ipar, iy1, iy2
  real                               rly1, rly2
! character*2 ccdmode
  integer       maxff, maxef, maxsw, maxlw, maxti, maxbu, npar, ntyp, maxsw100
  parameter   ( maxff=10, maxef=10, maxsw=3, maxlw=5, maxti=1, maxbu=1 )
  parameter   ( npar=9, ntyp=2, maxsw100 = 1 )
  real   ffpar(npar,maxff,ntyp), efpar(npar,maxef,ntyp)
  real   swpar(npar,maxsw,ntyp), lwpar(npar,maxlw,ntyp)
  real   tipar(npar,maxti,ntyp), bupar(npar,maxbu,ntyp)
  real   par(npar), swpar100(npar,maxsw100,ntyp)

! FullFrame: parameters for focus region (Y9) derived from  MS1229.2+6430, rev 0082
!                       for the other regions taken from ext. FF
  data ffpar/                                                           &
       0.5284, 1385.9, 0.1524, 158.155, 0.1762, 19.5666, 0.1435, 31.05, 10., &
       0.5284, 1385.9, 0.1524, 158.155, 0.1762, 19.5666, 0.1435, 31.93, 30., &
       0.5284, 1385.9, 0.1524, 158.155, 0.1762, 19.5666, 0.1435, 32.84, 50., &
       0.5354, 1394.1, 0.1342, 200.864, 0.1629, 30.6517, 0.1627, 33.77, 70., &
       0.5320, 1499.4, 0.1311, 222.169, 0.1601, 35.5882, 0.1747, 34.73, 90., &
       0.5334, 1725.7, 0.1222, 260.906, 0.1489, 42.7871, 0.1926, 35.71,110., &
       0.5553, 1184.2, 0.1230, 190.263, 0.1669, 31.6242, 0.1550, 36.73,130., &
       0.5650, 955.74, 0.1335, 163.263, 0.1682, 25.0995, 0.1349, 37.77,150., &
       0.5649, 943.63, 0.1404, 161.558, 0.1691, 24.2733, 0.1295, 38.84,170., &
       0.5331, 1329.,  0.1544, 180.6,   0.2007, 34.89,   0.1181, 40.00,190., &
       0.4101, 1356.3,-0.1121, 142.936,-0.1630, 19.8420,-0.1355, 31.05, 10., &
       0.4101, 1356.3,-0.1121, 142.936,-0.1630, 19.8420,-0.1355, 31.93, 30., &
       0.4101, 1356.3,-0.1121, 142.936,-0.1630, 19.8420,-0.1355, 32.84, 50., &
       0.4047, 1462.5,-0.0942, 193.856,-0.1428, 32.6295,-0.1629, 33.77, 70., &
       0.4015, 1484.8,-0.0849, 224.818,-0.1348, 38.8510,-0.1753, 34.73, 90., &
       0.4283, 2242.7,-0.1044, 259.620,-0.1286, 43.3512,-0.1936, 35.71,110., &
       0.3867, 987.81,-0.0955, 142.543,-0.1753, 15.0092,-0.1149, 36.73,130., &
       0.3812, 870.98,-0.0993, 139.727,-0.1687, 16.2656,-0.1134, 37.77,150., &
       0.3805, 825.41,-0.1073, 139.193,-0.1618, 19.0259,-0.1150, 38.84,170., &
       0.3988, 1359., -0.1016, 176.2,  -0.1836, 35.23,  -0.1198, 40.00,190./
 

! ExtendedFullFrame: parameters for 10 regions derived from merged Coma observations
  data efpar/                                                       &
       0.5284, 1385.9, 0.1524, 158.155, 0.1762, 19.5666, 0.1435, 31.05, 10., &
       0.5284, 1385.9, 0.1524, 158.155, 0.1762, 19.5666, 0.1435, 31.93, 30., &
       0.5284, 1385.9, 0.1524, 158.155, 0.1762, 19.5666, 0.1435, 32.84, 50., &
       0.5354, 1394.1, 0.1342, 200.864, 0.1629, 30.6517, 0.1627, 33.77, 70., &
       0.5320, 1499.4, 0.1311, 222.169, 0.1601, 35.5882, 0.1747, 34.73, 90., &
       0.5334, 1725.7, 0.1222, 260.906, 0.1489, 42.7871, 0.1926, 35.71,110., &
       0.5553, 1184.2, 0.1230, 190.263, 0.1669, 31.6242, 0.1550, 36.73,130., &
       0.5650, 955.74, 0.1335, 163.263, 0.1682, 25.0995, 0.1349, 37.77,150., &
       0.5649, 943.63, 0.1404, 161.558, 0.1691, 24.2733, 0.1295, 38.84,170., &
       0.5651, 963.51, 0.1460, 168.625, 0.1695, 22.5818, 0.1258, 40.00,190., &
       0.4101, 1356.3,-0.1121, 142.936,-0.1630, 19.8420,-0.1355, 31.05, 10., &
       0.4101, 1356.3,-0.1121, 142.936,-0.1630, 19.8420,-0.1355, 31.93, 30., &
       0.4101, 1356.3,-0.1121, 142.936,-0.1630, 19.8420,-0.1355, 32.84, 50., &
       0.4047, 1462.5,-0.0942, 193.856,-0.1428, 32.6295,-0.1629, 33.77, 70., &
       0.4015, 1484.8,-0.0849, 224.818,-0.1348, 38.8510,-0.1753, 34.73, 90., &
       0.4283, 2242.7,-0.1044, 259.620,-0.1286, 43.3512,-0.1936, 35.71,110., &
       0.3867, 987.81,-0.0955, 142.543,-0.1753, 15.0092,-0.1149, 36.73,130., &
       0.3812, 870.98,-0.0993, 139.727,-0.1687, 16.2656,-0.1134, 37.77,150., &
       0.3805, 825.41,-0.1073, 139.193,-0.1618, 19.0259,-0.1150, 38.84,170., &
       0.3823, 909.63,-0.1084, 156.678,-0.1599, 20.8893,-0.1210, 40.00,190./


! SmallWindow: parameters for focus region (Y9) derived from Mkn421 (ring), rev 0084
!              parameters for line 158 derived from Mkn421 (ring), rev 0084
!              then interpolated for Y6, Y7 and Y8
  data swpar/                                                           &
       0.5482,933.86, 0.1423,149.606, 0.1635,30.958, 0.1509,38.00,158., &
       0.5493,906.73, 0.1444,147.483, 0.1741,30.452, 0.1364,39.00,173., &
       0.5504,879.59, 0.1464,145.359, 0.1847,29.945, 0.1219,40.00,188., &
       0.3923,830.30,-0.1050,129.887,-0.1503,29.359,-0.1426,38.00,158., &
       0.3880,767.37,-0.1059,127.092,-0.1597,28.606,-0.1272,39.00,173., &
       0.3836,704.43,-0.1067,124.296,-0.1690,27.851,-0.1118,40.00,188./

  data swpar100/                                                        &
       0.5904,  84.3126, 0.08327, 476.665, 0.19736, 2422.91, 0.12886, 200.0, 190., &
       0.3595, 434.1561,-0.18502,  80.353,-0.07904, 2073.75,-0.09529, 200.0, 190./


! Large window: parameters for focus region (Y9) derived from Mkn 205, rev 0075
! parameters derived with CTE version of SAS 5.3
! all other Y areas are taken from eff
  data lwpar/                                                            &
       0.5334,1725.7, 0.1222,260.906, 0.1489,42.787, 0.1926,35.71,110.,  &
       0.5553,1184.2, 0.1230,190.263, 0.1669,31.624, 0.1550,36.73,130.,  &
       0.5650,955.74, 0.1335,163.263, 0.1682,25.100, 0.1349,37.77,150.,  &
       0.5649,943.63, 0.1404,161.558, 0.1691,24.273, 0.1295,38.84,170.,  &
       0.4930,1706.,  0.1723,189.7,   0.2082,34.30 , 0.1271,40.00,190.,  &
       0.4283,2242.7,-0.1044,259.620,-0.1286,43.351,-0.1936,35.71,110.,  &
       0.3867,987.81,-0.0955,142.543,-0.1753,15.009,-0.1149,36.73,130.,  &
       0.3812,870.98,-0.0993,139.727,-0.1687,16.266,-0.1134,37.77,150.,  &
       0.3805,825.41,-0.1073,139.193,-0.1618,19.026,-0.1150,38.84,170.,  &
       0.4250,1727., -0.1134,177.8,  -0.1885,32.89 ,-0.1245,40.00,190./


! Timing mode: parameters derived from GX13+1, rev 0057 (problems in rev 56-57?)
! data tipar/                                                            &
!      0.5805,6118.,  0.2479,965.9,   0.0538,135.2,  0.1100,80.00,190.,  &
!      0.3078,4585., -0.1425,968.2,  -0.0494,122.9, -0.1150,80.00,190./
! Timing mode: parameters derived from PKS2155-304, rev 0362
  data tipar/                                                            &
       0.5622,20891., 0.2307,880.1,   0.1007,126.2,  0.1033,65.00,190.,  &
       0.4137,18601.,-0.2168,857.6,  -0.0917,124.1, -0.1024,65.00,190./


! Burst mode: parameters derived from XTEJ1650, rev 0323
!             (consistent with Crab, rev 0234)
  data bupar/                                                           &
       0.6194,2173.,  0.1194,473.0,   0.1094,63.87,  0.1475,46.00,190., &
       0.3383,1963., -0.0982,420.9,  -0.0947,60.18, -0.1442,46.00,190./
  

  irun = irun+1
  rbinscal = 2.0
! if (ccdmode.eq.'FF') then
  if (calMode == 0) then
    iy1 = int(line/20.) + 1
    if (iy1.ge.maxff) iy1 = maxff - 1
    if (iy1.lt.1) iy1 = 1
    iy2 = iy1 + 1
    rly1 = 20.0*real(iy1) - 10.0
    rly2 = rly1 + 20.0
    do ipar = 1,npar
      par(ipar) = ffpar(ipar,iy2,pattyp) -     &
      (ffpar(ipar,iy2,pattyp)-ffpar(ipar,iy1,pattyp))*   &
      (rly2-line)/20.0
    enddo
! elseif (ccdmode.eq.'EF') then
  elseif (calMode == 1) then
    iy1 = int(line/20.) + 1
    if (iy1.ge.maxef) iy1 = maxef - 1
    if (iy1.lt.1) iy1 = 1
    iy2 = iy1 + 1
    rly1 = 20.0*real(iy1) - 10.0
    rly2 = rly1 + 20.0
    do ipar = 1,npar
      par(ipar) = efpar(ipar,iy2,pattyp) -     &
      (efpar(ipar,iy2,pattyp)-efpar(ipar,iy1,pattyp))*   &
      (rly2-line)/20.0
    enddo
! elseif (ccdmode.eq.'SW') then
  elseif (calMode == 3) then
    iy1 = int(line/20.) - 6
    if (iy1.ge.maxsw) iy1 = maxsw - 1
    if (iy1.lt.1) iy1 = 1
    iy2 = iy1 + 1
    rly1 = 20.0*real(iy1) + 130.0
    rly2 = rly1 + 20.0
    do ipar = 1,npar
      par(ipar) = swpar(ipar,iy2,pattyp) -     &
      (swpar(ipar,iy2,pattyp)-swpar(ipar,iy1,pattyp))*   &
      (rly2-line)/20.0
    enddo
! elseif (ccdmode.eq.'LW') then
  elseif (calMode == 2) then
    iy1 = int(line/20.) - 4
    if (iy1.ge.maxlw) iy1 = maxlw - 1
    if (iy1.lt.1) iy1 = 1
    iy2 = iy1 + 1
    rly1 = 20.0*real(iy1) + 90.0
    rly2 = rly1 + 20.0
    do ipar = 1,npar
      par(ipar) = lwpar(ipar,iy2,pattyp) -     &
      (lwpar(ipar,iy2,pattyp)-lwpar(ipar,iy1,pattyp))*   &
      (rly2-line)/20.0
    enddo
! elseif (ccdmode.eq.'TI') then
  elseif (calMode == 4) then
    iy1 = 1
    do ipar = 1,npar
      par(ipar) = tipar(ipar,iy1,pattyp)
    enddo
! elseif (ccdmode.eq.'BU') then
  elseif (calMode == 5) then
    iy1 = 1
    do ipar = 1,npar
      par(ipar) = bupar(ipar,iy1,pattyp)
    enddo
  elseif (calMode == 33) then
    iy1 = 1
    do ipar = 1,npar
      par(ipar) = swpar100(ipar,iy1,pattyp)
    enddo
    rbinscal = 2.0
  else
    do ipar = 1,7,2
      par(ipar) = 0.0
    enddo
    do ipar = 2,6,2
      par(ipar) = 1.0
    enddo
  endif
! par(8) = real(dthresh)/1.0014**(191.0-line)
  ndthresh = int(par(8))
  if (irun .le. 1) then
    call message(AppMsg,Verbose,' ')
    write(errstr,'(1x,a,t35,f9.4,i8)') 'Effective threshold: ',par(8),dthresh
    call message(AppMsg,Verbose)
  endif

!
! pattern fraction model: sum of 3 Lorentz functions + constant (= fraction at maximum energy)
!  LORE = LN/(1.+[ 2.*(X-LC)/LW ]^2)
!

if (calMode .ne. 33) then

! single-event pattern
  if (pattyp.eq.1) then
    do ibin = 1, ndthresh-1
      patfraction(ibin) = 1.0
    enddo
    do ibin = ndthresh, maxbin
      rbin = rbinscal*(real(ibin)-par(8))
!     if (calMode < 5) then
        patfraction(ibin) = min(1.0,      &
                            par(1)+par(3)/(1.0+(rbin/par(2))**2)   &
                           +par(5)/(1.0+(rbin/par(4))**2)   &
                           +par(7)/(1.0+(rbin/par(6))**2))
!     else
!       patfraction(ibin) = min(1.0,0.6915-0.125/0.69897*log10(real(ibin)/1000.)) 
!     endif
    enddo
! double-event pattern
  elseif (pattyp.eq.2) then
    do ibin = 1, ndthresh-1
      patfraction(ibin) = 0.0
    enddo
    do ibin = ndthresh, maxbin
      rbin = rbinscal*(real(ibin)-par(8))
!     if (calMode < 5) then
        patfraction(ibin) = max(0.0,      &
                            par(1)+par(3)/(1.0+(rbin/par(2))**2)   &
                           +par(5)/(1.0+(rbin/par(4))**2)   &
                           +par(7)/(1.0+(rbin/par(6))**2))
!     else
!       patfraction(ibin) = max(0.0,0.285+0.108/0.69897*log10(real(ibin)/1000.))
!     endif
    enddo
  endif

else

! rbinscal = 1.0

! single-event pattern
  if (pattyp.eq.1) then
    do ibin = 1, ndthresh-1
      patfraction(ibin) = 1.0
    enddo
    do ibin = ndthresh, maxbin
      rbin = rbinscal*(real(ibin)-par(8))
        patfraction(ibin) = min(1.0,      &
                            par(1)+par(3)/(1.0+(rbin/par(2))**2)   &
                           +par(5)/(1.0+(rbin/par(4))**2)   &
                           +par(7)/(1.0+(rbin/par(6))**2))
    enddo
! double-event pattern
  elseif (pattyp.eq.2) then
    do ibin = 1, ndthresh-1
      patfraction(ibin) = 0.0
    enddo
    do ibin = ndthresh, maxbin
      rbin = rbinscal*(real(ibin)-par(8))
        patfraction(ibin) = max(0.0,      &
                            par(1)+par(3)/(1.0+(rbin/par(2))**2)   &
                           +par(5)/(1.0+(rbin/par(4))**2)   &
                           +par(7)/(1.0+(rbin/par(6))**2))
    enddo
  endif


endif

end subroutine patmod



end module epatplot_mod
